#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "ptr.h"

static int ai[]={1, 0};
static int aj[]={0, 1};

int stlalg=EULER;
int stldir=FORWARD;


PTRData* crdp=NULL;

PTRData* CurrentData(){
    return crdp;
}

double ExtractV(PTRData* data, int i, int j, int k, int l)
{
    EnterFunc("ExtractV");
    if(data->nv<l){
        SetError(ERROR, E_NoEnoughLeng);
        LeaveFunc();  return 0.0;
    }else if(l<1){
        SetError(ERROR, E_CompNumReq);
        LeaveFunc();  return 0.0;
    }
    switch(data->dim){
      case 1:
        LeaveFunc();  return *(data->v[l-1]+i);
        break;
      case 2:
        LeaveFunc();  return *(data->v[l-1]+j*data->nx1+i);
        break;
      case 3:
        LeaveFunc();
        return *(data->v[l-1]+k*data->nx1*data->nx2+j*data->nx1+i);
        break;
      default:
        LeaveFunc();  return 0.0;
    }
}

int SearchNearestGrid(PTRData* field,
                      double x1, double x2, double x3,
                      int* i, int* j, int* k)
{
    int ista, iend, jsta, jend, ksta, kend;
    PTRData* fld;
    EnterFunc("SearchNearestGrid");
    *i=1;  *j=1;  *k=1;
    if(field){
        fld=field;
    }else if(!(fld=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return FALSE;
    }
    if(x1<fld->x1[0] || fld->x1[fld->nx1-1]<x1){
        LeaveFunc();  return FALSE;
    }
    if(1<fld->dim
       && (x2<fld->x2[0] || fld->x2[fld->nx2-1]<x2)){
        LeaveFunc();  return FALSE;
    }
    if(2<fld->dim
       && (x3<fld->x3[0] || fld->x3[fld->nx3-1]<x3)){
        LeaveFunc();  return FALSE;
    }
    ista=0;  iend=fld->nx1-1;
    *i=(ista+iend)/2;
    while(x1<fld->x1[*i] || fld->x1[*i+1]<x1){
        if(x1<fld->x1[*i]){
            iend=*i;
        }else if(fld->x1[*i+1]<x1){
            ista=*i;
        }
        *i=(ista+iend)/2;
    }
    if(fld->dim==1){
        LeaveFunc();  return TRUE;
    }
    jsta=0;  jend=fld->nx2-1;
    *j=(jsta+jend)/2;
    while(x2<fld->x2[*j] || fld->x2[*j+1]<x2){
        if(x2<fld->x2[*j]){
            jend=*j;
        }else if(fld->x2[*j+1]<x2){
            jsta=*j;
        }
        *j=(jsta+jend)/2;
    }
    if(fld->dim==2){
        LeaveFunc();  return TRUE;
    }
    ksta=0;  kend=fld->nx3-1;
    *k=(ksta+kend)/2;
    while(x3<fld->x3[*k] || fld->x3[*k+1]<x3){
        if(x3<fld->x3[*k]){
            kend=*k;
        }else if(fld->x3[*k+1]<x3){
            ksta=*k;
        }
        *k=(ksta+kend)/2;
    }
    LeaveFunc();  return TRUE;
}

int InterpolateVector(PTRData* field,
                      double x1, double x2, double x3,
                      int nv, int vi[], double vt[])
{
    int i, j, k, l, m;
    PTRData* fld;
    EnterFunc("InterpolateVector");
    if(field){
        fld=field;
    }else if(!(fld=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return FALSE;
    }
    for(m=0; m<nv; m++){
        if(vi[m]<1){
            SetError(ERROR, E_CompNumReq);
            LeaveFunc();  return FALSE;
        }else if(fld->nv<vi[m]){
            SetError(ERROR, E_NoEnoughLeng);
            LeaveFunc();  return FALSE;
        }
    }
    if(fld->type!=GRID){
        SetError(ERROR, E_BadDataType);
        LeaveFunc();  return FALSE;
    }
    if(!(SearchNearestGrid(fld, x1, x2, x3, &i, &j, &k))){
        LeaveFunc();  return FALSE;
    }
    if(fld->dim==1){
        double dx, dx10, dx11;
        dx10=x1-fld->x1[i];  dx11=fld->x1[i+1]-x1;
        dx=1.0/(fld->x1[i+1]-fld->x1[i]);
        for(m=0; m<nv; m++){
            l=vi[m];
            vt[m]=dx*( dx10*ExtractV(fld, i+1, j, k, l)
                      +dx11*ExtractV(fld, i  , j, k, l));
        }
    }else if(fld->dim==2){
        double dx, dx10, dx11, dx20, dx21;
        dx10=x1-fld->x1[i];  dx11=fld->x1[i+1]-x1;
        dx20=x2-fld->x2[j];  dx21=fld->x2[j+1]-x2;
        dx=1.0/(fld->x1[i+1]-fld->x1[i])/(fld->x2[j+1]-fld->x2[j]);
        for(m=0; m<nv; m++){
            l=vi[m];
            vt[m]=dx*( dx10*dx20*ExtractV(fld, i+1, j+1, k, l)
                      +dx11*dx20*ExtractV(fld, i  , j+1, k, l)
                      +dx10*dx21*ExtractV(fld, i+1, j  , k, l)
                      +dx11*dx21*ExtractV(fld, i  , j  , k, l));
        }
    }else if(fld->dim==3){
        double dx, dx10, dx11, dx20, dx21, dx30, dx31;
        dx10=x1-fld->x1[i];  dx11=fld->x1[i+1]-x1;
        dx20=x2-fld->x2[j];  dx21=fld->x2[j+1]-x2;
        dx30=x3-fld->x3[k];  dx31=fld->x3[k+1]-x3;
        dx=1.0/(fld->x1[i+1]-fld->x1[i])
                /(fld->x2[j+1]-fld->x2[j])
                        /(fld->x3[k+1]-fld->x3[k]);
        for(m=0; m<nv; m++){
        l=vi[m];
        vt[m]=dx*( dx10*dx20*dx30*ExtractV(fld, i+1, j+1, k+1, l)
                  +dx11*dx20*dx30*ExtractV(fld, i  , j+1, k+1, l)
                  +dx10*dx21*dx30*ExtractV(fld, i+1, j  , k+1, l)
                  +dx11*dx21*dx30*ExtractV(fld, i  , j  , k+1, l)
                  +dx10*dx20*dx31*ExtractV(fld, i+1, j+1, k  , l)
                  +dx11*dx20*dx31*ExtractV(fld, i  , j+1, k  , l)
                  +dx10*dx21*dx31*ExtractV(fld, i+1, j  , k  , l)
                  +dx11*dx21*dx31*ExtractV(fld, i  , j  , k  , l));
        }
    }
    LeaveFunc();  return TRUE;
}

/* create data */
PTRData* CreateGridData(int dim, int nx1, int nx2, int nx3, int nv,
                        double* x1, double* x2,
                        double* x3, double* v[])
{
    PTRData* dp;
    dp=(PTRData*)malloc(sizeof(PTRData));
    dp->type=GRID;
    dp->nx1=1;  dp->nx2=1;  dp->nx3=1;
    dp->dim=dim;
    dp->nv=nv;
    dp->x1=NULL;  dp->x2=NULL;  dp->x3=NULL;
    dp->v =NULL;
    if(1<=dim){
        dp->nx1=nx1;
        dp->x1=(double*)malloc(nx1*sizeof(double));
        memcpy(dp->x1, x1, nx1*sizeof(double));
    }
    if(2<=dim){
        dp->nx2=nx2;
        dp->x2=(double*)malloc(nx2*sizeof(double));
        memcpy(dp->x2, x2, nx2*sizeof(double));
    }
    if(3<=dim){
        dp->nx3=nx3;
        dp->x3=(double*)malloc(nx3*sizeof(double));
        memcpy(dp->x3, x3, nx3*sizeof(double));
    }
    if(1<=nv){
        int i, n;
        n=nx1*nx2*nx3;
        dp->v=(double**)malloc(nv*sizeof(double*));
        for(i=0; i<nv; i++){
            dp->v[i]=(double*)malloc(n*sizeof(double));
            memcpy(dp->v[i], v[i], n*sizeof(double));
        }
    }
    return dp;
}
void ReadGridData(int dim, int nx1, int nx2, int nx3, int nv,
                  double* x1, double* x2, double* x3, double* v[])
{
    if(crdp) FreeData(crdp);
    crdp=CreateGridData(dim, nx1, nx2, nx3, nv, x1, x2, x3, v);
}

PTRData* CreateSeqData(int dim, int nx, int nv,
                       double* x1, double* x2,
                       double* x3, double* v[])
{
    PTRData* dp;
    dp=(PTRData*)malloc(sizeof(PTRData));
    dp->type=SEQ;
    dp->nx1=nx;  dp->nx2=nx;  dp->nx3=nx;
    dp->dim=dim;
    dp->nv=nv;
    dp->x1=NULL;  dp->x2=NULL;  dp->x3=NULL;
    dp->v =NULL;
    if(1<=dim){
        dp->x1=(double*)malloc(nx*sizeof(double));
        memcpy(dp->x1, x1, nx*sizeof(double));
    }
    if(2<=dim){
        dp->x2=(double*)malloc(nx*sizeof(double));
        memcpy(dp->x2, x2, nx*sizeof(double));
    }
    if(3<=dim){
        dp->x3=(double*)malloc(nx*sizeof(double));
        memcpy(dp->x3, x3, nx*sizeof(double));
    }
    if(1<=nv){
        int i;
        dp->v=(double**)malloc(nv*sizeof(double*));
        for(i=0; i<nv; i++){
            dp->v[i]=(double*)malloc(nx*sizeof(double));
            memcpy(dp->v[i], v[i], nx*sizeof(double));
        }
    }
    return dp;
}
void ReadSeqData(int dim, int nx, int nv,
                 double* x1, double* x2,
                 double* x3, double* v[])
{
    if(crdp) FreeData(crdp);
    crdp=CreateSeqData(dim, nx, nv, x1, x2, x3, v);
}

void ReadData(int nx, double* x1, double* v){
    ReadSeqData(1, nx, 1, x1, NULL, NULL, &v);
}

PTRData* DuplicateData(PTRData* dp)
{
    PTRData* dp1;
    dp1=(PTRData*)malloc(sizeof(PTRData));
    *dp1=*dp;
    if(dp1->x1){
        dp1->x1=(double*)malloc(dp->nx1*sizeof(double));
        memcpy(dp1->x1, dp->x1, dp->nx1*sizeof(double));
    }
    if(dp1->x2){
        dp1->x2=(double*)malloc(dp->nx2*sizeof(double));
        memcpy(dp1->x2, dp->x2, dp->nx2*sizeof(double));
    }
    if(dp1->x3){
        dp1->x3=(double*)malloc(dp->nx3*sizeof(double));
        memcpy(dp1->x3, dp->x3, dp->nx3*sizeof(double));
    }
    if(dp1->v){
        int i, n;
        if(dp->type==GRID){
            n=dp->nx1*dp->nx2*dp->nx3;
        }else if(dp->type==SEQ){
            n=dp->nx1;
        }
        dp1->v=(double**)malloc(dp->nv*sizeof(double*));
        for(i=0; i<dp->nv; i++){
            dp1->v[i]=(double*)malloc(n*sizeof(double));
            memcpy(dp1->v[i], dp->v[i], n*sizeof(double));
        }
    }
    return dp1;
}

void SaveGridData(FILE* fp, PTRData* dp)
{
    int buf[6];
    buf[0]=dp->type;
    buf[1]=dp->nx1;  buf[2]=dp->nx2;  buf[3]=dp->nx3;
    buf[4]=dp->dim;  buf[5]=dp->nv;
    fwrite(buf, sizeof(int), 6, fp);
    if(1<=dp->dim)
            fwrite(dp->x1, sizeof(double), dp->nx1, fp);
    if(2<=dp->dim)
            fwrite(dp->x2, sizeof(double), dp->nx2, fp);
    if(3<=dp->dim)
            fwrite(dp->x3, sizeof(double), dp->nx3, fp);
    if(1<=dp->nv){
        int i, leng=dp->nx1*dp->nx2*dp->nx3;
        for(i=0; i<dp->nv; i++){
            fwrite(dp->v[i], sizeof(double), leng, fp);
        }
    }
}
void SaveSeqData(FILE* fp, PTRData* dp)
{
    int buf[6];
    buf[0]=dp->type;
    buf[1]=dp->nx1;  buf[2]=dp->nx2;  buf[3]=dp->nx3;
    buf[4]=dp->dim;  buf[5]=dp->nv;
    fwrite(buf, sizeof(int), 6, fp);
    if(1<=dp->dim)
            fwrite(dp->x1, sizeof(double), dp->nx1, fp);
    if(2<=dp->dim)
            fwrite(dp->x2, sizeof(double), dp->nx2, fp);
    if(3<=dp->dim)
            fwrite(dp->x3, sizeof(double), dp->nx3, fp);
    if(1<=dp->nv){
        int i;
        for(i=0; i<dp->nv; i++){
            fwrite(dp->v[i], sizeof(double), dp->nx1, fp);
        }
    }
}
void SaveData(char* fname, PTRData* data)
{
    FILE* fp;
    PTRData* dp;
    EnterFunc("SaveData");
    if(data){
        dp=data;
    }else if(!(dp=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return;
    }
    if(!OverWrite(fname)) return;
    if(!(fp=fopen(fname, "w"))){
        char msg[MAX_STR];
        sprintf(msg, E_FileError, fname);
        SetError(ERROR, msg);
        LeaveFunc();  return;
    }
    if(dp->type==GRID){
        SaveGridData(fp, dp);
    }else if(dp->type==SEQ){
        SaveSeqData(fp, dp);
    }else{
        SetError(ERROR, E_BadDataType);
        LeaveFunc();  return;
    }
    LeaveFunc();
}
void SaveDataFP(FILE* fp, PTRData* data)
{
    PTRData* dp;
    EnterFunc("SaveDataFP");
    if(data){
        dp=data;
    }else if(!(dp=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return;
    }
    if(dp->type==GRID){
        SaveGridData(fp, dp);
    }else if(dp->type==SEQ){
        SaveSeqData(fp, dp);
    }else{
        SetError(ERROR, E_BadDataType);
        LeaveFunc();  return;
    }
    LeaveFunc();
}

PTRData* RestoreGridData(FILE* fp)
{
    int buf[5];
    PTRData* dp;
    dp=(PTRData*)malloc(sizeof(PTRData));
    dp->type=GRID;
    fread(buf, sizeof(int), 5, fp);
    dp->nx1=buf[0];  dp->nx2=buf[1];  dp->nx3=buf[2];
    dp->dim=buf[3];  dp->nv =buf[4];
    dp->x1=NULL;  dp->x2=NULL;  dp->x3=NULL;
    dp->v =NULL;
    if(1<=dp->dim){
        dp->x1=(double*)malloc(dp->nx1*sizeof(double));
        fread(dp->x1, sizeof(double), dp->nx1, fp);
    }
    if(2<=dp->dim){
        dp->x2=(double*)malloc(dp->nx2*sizeof(double));
        fread(dp->x2, sizeof(double), dp->nx2, fp);
    }
    if(3<=dp->dim){
        dp->x3=(double*)malloc(dp->nx3*sizeof(double));
        fread(dp->x3, sizeof(double), dp->nx3, fp);
    }
    if(1<=dp->nv){
        int i, leng=dp->nx1*dp->nx2*dp->nx3;
        dp->v=(double**)malloc(dp->nv*sizeof(double*));
        for(i=0; i<dp->nv; i++){
            dp->v[i]=(double*)malloc(leng*sizeof(double));
            fread(dp->v[i], sizeof(double), leng, fp);
        }
    }
    return dp;
}
PTRData* RestoreSeqData(FILE* fp)
{
    int buf[5];
    PTRData* dp;
    dp=(PTRData*)malloc(sizeof(PTRData));
    dp->type=SEQ;
    fread(buf, sizeof(int), 5, fp);
    dp->nx1=buf[0];  dp->nx2=buf[1];  dp->nx3=buf[2];
    dp->dim=buf[3];  dp->nv =buf[4];
    dp->x1=NULL;  dp->x2=NULL;  dp->x3=NULL;
    dp->v =NULL;
    if(1<=dp->dim){
        dp->x1=(double*)malloc(dp->nx1*sizeof(double));
        fread(dp->x1, sizeof(double), dp->nx1, fp);
    }
    if(2<=dp->dim){
        dp->x2=(double*)malloc(dp->nx2*sizeof(double));
        fread(dp->x2, sizeof(double), dp->nx2, fp);
    }
    if(3<=dp->dim){
        dp->x3=(double*)malloc(dp->nx3*sizeof(double));
        fread(dp->x3, sizeof(double), dp->nx3, fp);
    }
    if(1<=dp->nv){
        int i;
        dp->v=(double**)malloc(dp->nv*sizeof(double*));
        for(i=0; i<dp->nv; i++){
            dp->v[i]=(double*)malloc(dp->nx1*sizeof(double));
            fread(dp->v[i], sizeof(double), dp->nx1, fp);
        }
    }
    return dp;
}
PTRData* RestoreData(char* fname)
{
    int type;
    FILE* fp;
    PTRData* dp;
    EnterFunc("RestoreData");
    if(!(fp=fopen(fname, "r"))){
        char msg[MAX_STR];
        sprintf(msg, E_FileError, fname);
        SetError(ERROR, msg);
        LeaveFunc();  return NULL;
    }
    fread(&type, sizeof(int), 1, fp);
    if(type==GRID){
        dp=RestoreGridData(fp);
    }else if(type==SEQ){
        dp=RestoreSeqData(fp);
    }else{
        SetError(ERROR, E_BadDataType);
        LeaveFunc();  return NULL;
    }
    LeaveFunc();  return dp;
}
PTRData* RestoreDataFP(FILE* fp)
{
    int type;
    PTRData* dp;
    EnterFunc("RestoreDataFP");
    fread(&type, sizeof(int), 1, fp);
    if(type==GRID){
        dp=RestoreGridData(fp);
    }else if(type==SEQ){
        dp=RestoreSeqData(fp);
    }else{
        SetError(ERROR, E_BadDataType);
        LeaveFunc();  return NULL;
    }
    LeaveFunc();  return dp;
}

void LoadData(char* fname)
{
    PTRData* dp;
    EnterFunc("LoadData");
    if((dp=RestoreData(fname))){
        if(crdp) FreeData(crdp);
        crdp=dp;
    }
    LeaveFunc();
}



/* filters */

/* exrtact data */
PTRData* ExtractSeqDataI(PTRData* data, int j, int k)
{
    int i, l, lj, lk, nx;
    PTRData* dt;
    PTRData* ret;
    EnterFunc("ExtractSeqDataI");
    if(data){
        dt=data;
    }else if(!(dt=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(dt->dim<=1) return NULL;
    nx=dt->nx1;
    ret=(PTRData*)malloc(sizeof(PTRData));
    ret->type=SEQ;
    ret->nx1=nx;  ret->nx2=nx;  ret->nx3=nx;
    ret->dim=dt->dim;
    ret->nv=dt->nv;
    ret->x1=NULL;  ret->x2=NULL;  ret->x3=NULL;
    ret->v =NULL;
    ret->x1=(double*)malloc(nx*sizeof(double));
    ret->x2=(double*)malloc(nx*sizeof(double));
    if(3<=dt->dim) ret->x3=(double*)malloc(nx*sizeof(double));
    if(1<=dt->nv){
        ret->v=(double**)malloc(dt->nv*sizeof(double*));
        for(l=0; l<dt->nv; l++)
                ret->v[l]=(double*)malloc(nx*sizeof(double));
    }
    for(i=0; i<nx; i++){
        ret->x1[i]=dt->x1[i];
        ret->x2[i]=dt->x2[j];
        if(ret->x3) ret->x3[i]=dt->x3[k];
    }
    lk=dt->nx2*dt->nx1;
    lj=dt->nx1;
    if(ret->v){
        for(l=0; l<dt->nv; l++){
            for(i=0; i<nx; i++)
                    ret->v[l][i]=dt->v[l][k*lk+j*lj+i];
        }
    }
    return ret;
}
PTRData* ExtractSeqDataJ(PTRData* data, int i, int k)
{
    int j, l, lj, lk, nx;
    PTRData* dt;
    PTRData* ret;
    EnterFunc("ExtractSeqDataJ");
    if(data){
        dt=data;
    }else if(!(dt=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(dt->dim<=1) return NULL;
    nx=dt->nx2;
    ret=(PTRData*)malloc(sizeof(PTRData));
    ret->type=SEQ;
    ret->nx1=nx;  ret->nx2=nx;  ret->nx3=nx;
    ret->dim=dt->dim;
    ret->nv=dt->nv;
    ret->x1=NULL;  ret->x2=NULL;  ret->x3=NULL;
    ret->v =NULL;
    ret->x1=(double*)malloc(nx*sizeof(double));
    ret->x2=(double*)malloc(nx*sizeof(double));
    if(3<=dt->dim) ret->x3=(double*)malloc(nx*sizeof(double));
    if(1<=dt->nv){
        ret->v=(double**)malloc(dt->nv*sizeof(double*));
        for(l=0; l<dt->nv; l++)
                ret->v[l]=(double*)malloc(nx*sizeof(double));
    }
    for(j=0; j<nx; j++){
        ret->x1[j]=dt->x1[i];
        ret->x2[j]=dt->x2[j];
        if(ret->x3) ret->x3[j]=dt->x3[k];
    }
    lk=dt->nx2*dt->nx1;
    lj=dt->nx1;
    if(ret->v){
        for(l=0; l<dt->nv; l++){
            for(j=0; j<nx; j++)
                    ret->v[l][j]=dt->v[l][k*lk+j*lj+i];
        }
    }
    return ret;
}
PTRData* ExtractSeqDataK(PTRData* data, int i, int j)
{
    int k, l, lj, lk, nx;
    PTRData* dt;
    PTRData* ret;
    EnterFunc("ExtractSeqDataK");
    if(data){
        dt=data;
    }else if(!(dt=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(dt->dim<=2) return NULL;
    nx=dt->nx3;
    ret=(PTRData*)malloc(sizeof(PTRData));
    ret->type=SEQ;
    ret->nx1=nx;  ret->nx2=nx;  ret->nx3=nx;
    ret->dim=dt->dim;
    ret->nv=dt->nv;
    ret->x1=NULL;  ret->x2=NULL;  ret->x3=NULL;
    ret->v =NULL;
    ret->x1=(double*)malloc(nx*sizeof(double));
    ret->x2=(double*)malloc(nx*sizeof(double));
    ret->x3=(double*)malloc(nx*sizeof(double));
    if(1<=dt->nv){
        ret->v=(double**)malloc(dt->nv*sizeof(double*));
        for(l=0; l<dt->nv; l++)
                ret->v[l]=(double*)malloc(nx*sizeof(double));
    }
    for(k=0; k<nx; k++){
        ret->x1[k]=dt->x1[i];
        ret->x2[k]=dt->x2[j];
        ret->x3[k]=dt->x3[k];
    }
    lk=dt->nx2*dt->nx1;
    lj=dt->nx1;
    if(ret->v){
        for(l=0; l<dt->nv; l++){
            for(k=0; k<nx; k++)
                    ret->v[l][k]=dt->v[l][k*lk+j*lj+i];
        }
    }
    return ret;
}

/* down size */

PTRData* DownSizeSeqData(PTRData* data, int is)
{
    int i, l, nx=0;
    PTRData* dt;
    PTRData* ret;
    EnterFunc("DownSizeSeqData");
    if(data){
        dt=data;
    }else if(!(dt=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(is<1){
        SetError(ERROR, E_BadValue);
        LeaveFunc();  return NULL;
    }else if(is==1){
        return DuplicateData(dt);
    }
    for(i=0; i<dt->nx1; i+=is) nx++;
    ret=(PTRData*)malloc(sizeof(PTRData));
    ret->type=SEQ;
    ret->nx1=nx;  ret->nx2=nx;  ret->nx3=nx;
    ret->dim=dt->dim;
    ret->nv=dt->nv;
    ret->x1=NULL;  ret->x2=NULL;  ret->x3=NULL;
    ret->v =NULL;
    ret->x1=(double*)malloc(nx*sizeof(double));
    if(2<=dt->dim && dt->x2)
            ret->x2=(double*)malloc(nx*sizeof(double));
    if(3<=dt->dim && dt->x3)
            ret->x3=(double*)malloc(nx*sizeof(double));
    if(1<=dt->nv){
        ret->v=(double**)malloc(dt->nv*sizeof(double*));
        for(l=0; l<dt->nv; l++)
                ret->v[l]=(double*)malloc(nx*sizeof(double));
    }
    for(i=0; i<nx; i++) ret->x1[i]=dt->x1[is*i];
    if(ret->x2) for(i=0; i<nx; i++) ret->x2[i]=dt->x2[is*i];
    if(ret->x3) for(i=0; i<nx; i++) ret->x3[i]=dt->x3[is*i];
    if(ret->v){
        for(l=0; l<dt->nv; l++){
            for(i=0; i<nx; i++) ret->v[l][i]=dt->v[l][is*i];
        }
    }
    return ret;
}

PTRData* DownSizeData(PTRData* data, int is, int js, int ks)
{
    PTRData* dt;
    EnterFunc("DownSizeData");
    if(data){
        dt=data;
    }else if(!(dt=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(dt->type==SEQ){
        return DownSizeSeqData(dt, is);
    }
}


/* line contour */
int SearchStartPointK(PTRData* field, int l, double val,
                      char* map, int* i, int* j, int k,
                      int* d, int* e)
{
    double a, b;
    *j=0;  *d=0;  *e=0;
    for(*i=0; *i<field->nx1-1; (*i)++){
        a=ExtractV(field, *i, *j, k, l);
        if(!*(map+*j*2*field->nx1+*i*2+*d)){
            b=ExtractV(field, *i+ai[*d], *j+aj[*d], k, l);
            if((a-val)*(b-val)<=0 && b!=a) return TRUE;
            *(map+*j*2*field->nx1+*i*2+*d)=TRUE;
        }
    }
    *i=field->nx1-1;  *d=1;  *e=1;
    for(*j=0; *j<field->nx2-1; (*j)++){
        a=ExtractV(field, *i, *j, k, l);
        if(!*(map+*j*2*field->nx1+*i*2+*d)){
            b=ExtractV(field, *i+ai[*d], *j+aj[*d], k, l);
            if((a-val)*(b-val)<=0 && b!=a) return TRUE;
            *(map+*j*2*field->nx1+*i*2+*d)=TRUE;
        }
    }
    *j=field->nx2-1;  *d=0;  *e=1;
    for(*i=field->nx1-2; *i>=0; (*i)--){
        a=ExtractV(field, *i, *j, k, l);
        if(!*(map+*j*2*field->nx1+*i*2+*d)){
            b=ExtractV(field, *i+ai[*d], *j+aj[*d], k, l);
            if((a-val)*(b-val)<=0 && b!=a) return TRUE;
            *(map+*j*2*field->nx1+*i*2+*d)=TRUE;
        }
    }
    *i=0;  *d=1;  *e=0;
    for(*j=field->nx2-2; *j>=0; (*j)--){
        a=ExtractV(field, *i, *j, k, l);
        if(!*(map+*j*2*field->nx1+*i*2+*d)){
            b=ExtractV(field, *i+ai[*d], *j+aj[*d], k, l);
            if((a-val)*(b-val)<=0 && b!=a) return TRUE;
            *(map+*j*2*field->nx1+*i*2+*d)=TRUE;
        }
    }
    for(*j=0; *j<field->nx2-1; (*j)++){
        *e=0;
        for(*i=0; *i<field->nx1-1; (*i)++){
            for(*d=0; *d<2; (*d)++){
                a=ExtractV(field, *i, *j, k, l);
                if(!*(map+*j*2*field->nx1+*i*2+*d)){
                    b=ExtractV(field, *i+ai[*d], *j+aj[*d], k, l);
                    if((a-val)*(b-val)<=0 && b!=a) return TRUE;
                    *(map+*j*2*field->nx1+*i*2+*d)=TRUE;
                }
            }
        }
    }
    return FALSE;
}
int SearchStartPointI(PTRData* field, int l, double val,
                      char* map, int i, int* j, int* k,
                      int* d, int* e)
{
    double a, b;
    *k=0;  *d=0;  *e=0;
    for(*j=0; *j<field->nx2-1; (*j)++){
        a=ExtractV(field, i, *j, *k, l);
        if(!*(map+*k*2*field->nx2+*j*2+*d)){
            b=ExtractV(field, i, *j+ai[*d], *k+aj[*d], l);
            if((a-val)*(b-val)<=0 && b!=a) return TRUE;
            *(map+*k*2*field->nx2+*j*2+*d)=TRUE;
        }
    }
    *j=field->nx2-1;  *d=1;  *e=1;
    for(*k=0; *k<field->nx3-1; (*k)++){
        a=ExtractV(field, i, *j, *k, l);
        if(!*(map+*k*2*field->nx2+*j*2+*d)){
            b=ExtractV(field, i, *j+ai[*d], *k+aj[*d], l);
            if((a-val)*(b-val)<=0 && b!=a) return TRUE;
            *(map+*k*2*field->nx2+*j*2+*d)=TRUE;
        }
    }
    *k=field->nx3-1;  *d=0;  *e=1;
    for(*j=field->nx2-2; *j>=0; (*j)--){
        a=ExtractV(field, i, *j, *k, l);
        if(!*(map+*k*2*field->nx2+*j*2+*d)){
            b=ExtractV(field, i, *j+ai[*d], *k+aj[*d], l);
            if((a-val)*(b-val)<=0 && b!=a) return TRUE;
            *(map+*k*2*field->nx2+*j*2+*d)=TRUE;
        }
    }
    *j=0;  *d=1;  *e=0;
    for(*k=field->nx3-2; *k>=0; (*k)--){
        a=ExtractV(field, i, *j, *k, l);
        if(!*(map+*k*2*field->nx2+*j*2+*d)){
            b=ExtractV(field, i, *j+ai[*d], *k+aj[*d], l);
            if((a-val)*(b-val)<=0 && b!=a) return TRUE;
            *(map+*k*2*field->nx2+*j*2+*d)=TRUE;
        }
    }
    for(*k=0; *k<field->nx3-1; (*k)++){
        *e=0;
        for(*j=0; *j<field->nx2-1; (*j)++){
            for(*d=0; *d<2; (*d)++){
                a=ExtractV(field, i, *j, *k, l);
                if(!*(map+*k*2*field->nx2+*j*2+*d)){
                    b=ExtractV(field, i, *j+ai[*d], *k+aj[*d], l);
                    if((a-val)*(b-val)<=0 && b!=a) return TRUE;
                    *(map+*k*2*field->nx2+*j*2+*d)=TRUE;
                }
            }
        }
    }
    return FALSE;
}
int SearchStartPointJ(PTRData* field, int l, double val,
                      char* map, int* i, int j, int* k,
                      int* d, int* e)
{
    double a, b;
    *i=0;  *d=0;  *e=0;
    for(*k=0; *k<field->nx3-1; (*k)++){
        a=ExtractV(field, *i, j, *k, l);
        if(!*(map+*i*2*field->nx3+*k*2+*d)){
            b=ExtractV(field, *i+aj[*d], j, *k+ai[*d], l);
            if((a-val)*(b-val)<=0 && b!=a) return TRUE;
            *(map+*i*2*field->nx3+*k*2+*d)=TRUE;
        }
    }
    *k=field->nx3-1;  *d=1;  *e=1;
    for(*i=0; *i<field->nx1-1; (*i)++){
        a=ExtractV(field, *i, j, *k, l);
        if(!*(map+*i*2*field->nx3+*k*2+*d)){
            b=ExtractV(field, *i+aj[*d], j, *k+ai[*d], l);
            if((a-val)*(b-val)<=0 && b!=a) return TRUE;
            *(map+*i*2*field->nx3+*k*2+*d)=TRUE;
        }
    }
    *i=field->nx1-1;  *d=0;  *e=1;
    for(*k=field->nx3-2; *k>=0; (*k)--){
        a=ExtractV(field, *i, j, *k, l);
        if(!*(map+*i*2*field->nx3+*k*2+*d)){
            b=ExtractV(field, *i+aj[*d], j, *k+ai[*d], l);
            if((a-val)*(b-val)<=0 && b!=a) return TRUE;
            *(map+*i*2*field->nx3+*k*2+*d)=TRUE;
        }
    }
    *k=0;  *d=1;  *e=0;
    for(*i=field->nx1-2; *i>=0; (*i)--){
        a=ExtractV(field, *i, j, *k, l);
        if(!*(map+*i*2*field->nx3+*k*2+*d)){
            b=ExtractV(field, *i+aj[*d], j, *k+ai[*d], l);
            if((a-val)*(b-val)<=0 && b!=a) return TRUE;
            *(map+*i*2*field->nx3+*k*2+*d)=TRUE;
        }
    }
    for(*i=0; *i<field->nx1-1; (*i)++){
        *e=0;
        for(*k=0; *k<field->nx3-1; (*k)++){
            for(*d=0; *d<2; (*d)++){
                a=ExtractV(field, *i, j, *k, l);
                if(!*(map+*i*2*field->nx3+*k*2+*d)){
                    b=ExtractV(field, *i+aj[*d], j, *k+ai[*d], l);
                    if((a-val)*(b-val)<=0 && b!=a) return TRUE;
                    *(map+*i*2*field->nx3+*k*2+*d)=TRUE;
                }
            }
        }
    }
    return FALSE;
}
int NextPointK(PTRData* field, int l, double val, char* map,
               int* i, int* j, int k, int* d, int* e,
               double* x1, double* x2)
{
    int m, i0, j0, i1, j1, d0, e0;
    double a, b;
    static int dd[][3]={{1, 0, 1}, {0, 1, 0}};
    static int ee[][2][3]
            ={{{0, 0, 1}, {1, 1, 0}}, {{1, 0, 0}, {0, 1, 1}}};
    static int ii[][2][3]
            ={{{1, 0, 0}, {0, 0, 1}}, {{0, 1, 0},{-1,-1,-1}}};
    static int jj[][2][3]
            ={{{0, 1, 0},{-1,-1,-1}}, {{0, 0, 1}, {1, 0, 0}}};
    d0=*d;  i0=*i;  j0=*j;
    for(m=0; m<3; m++){
        d0=dd[*d][m];
        e0=ee[*d][*e][m];
        i0=*i+ii[*d][*e][m];  j0=*j+jj[*d][*e][m];
        i1=i0+ai[d0];  j1=j0+aj[d0];
        if(*(map+j0*2*field->nx1+i0*2+d0)
           || i0<0 || field->nx1<=i0 || j0<0 || field->nx2<=j0
           || i1<0 || field->nx1<=i1 || j1<0 || field->nx2<=j1){
            continue;
        }
        *(map+j0*2*field->nx1+i0*2+d0)=TRUE;
        a=ExtractV(field, i0, j0, k, l);
        b=ExtractV(field, i1, j1, k, l);
        if((a-val)*(b-val)<=0 && b!=a){
            double x10, x11, x20, x21;
            x10=field->x1[i0];  x11=field->x1[i1];
            x20=field->x2[j0];  x21=field->x2[j1];
            *x1=x10+(val-a)/(b-a)*(x11-x10);
            *x2=x20+(val-a)/(b-a)*(x21-x20);
            *i=i0;  *j=j0;  *d=d0;  *e=e0;
            return TRUE;
        }
    }
    *(map+*j*2*field->nx1+*i*2+*d)=TRUE;
    return FALSE;
}
int NextPointI(PTRData* field, int l, double val, char* map,
               int i, int* j, int* k, int* d, int* e,
               double* x2, double* x3)
{
    int m, j0, k0, j1, k1, d0, e0;
    double a, b;
    static int dd[][3]={{1, 0, 1}, {0, 1, 0}};
    static int ee[][2][3]
            ={{{0, 0, 1}, {1, 1, 0}}, {{1, 0, 0}, {0, 1, 1}}};
    static int ii[][2][3]
            ={{{1, 0, 0}, {0, 0, 1}}, {{0, 1, 0},{-1,-1,-1}}};
    static int jj[][2][3]
            ={{{0, 1, 0},{-1,-1,-1}}, {{0, 0, 1}, {1, 0, 0}}};
    d0=*d;  j0=*j;  k0=*k;
    for(m=0; m<3; m++){
        d0=dd[*d][m];
        e0=ee[*d][*e][m];
        j0=*j+ii[*d][*e][m];  k0=*k+jj[*d][*e][m];
        j1=j0+ai[d0];  k1=k0+aj[d0];
        if(*(map+k0*2*field->nx2+j0*2+d0)
           || j0<0 || field->nx2<=j0 || k0<0 || field->nx3<=k0
           || j1<0 || field->nx2<=j1 || k1<0 || field->nx3<=k1){
            continue;
        }
        *(map+k0*2*field->nx2+j0*2+d0)=TRUE;
        a=ExtractV(field, i, j0, k0, l);
        b=ExtractV(field, i, j1, k1, l);
        if((a-val)*(b-val)<=0 && b!=a){
            double x20, x21, x30, x31;
            x20=field->x2[j0];  x21=field->x2[j1];
            x30=field->x3[k0];  x31=field->x3[k1];
            *x2=x20+(val-a)/(b-a)*(x21-x20);
            *x3=x30+(val-a)/(b-a)*(x31-x30);
            *j=j0;  *k=k0;  *d=d0;  *e=e0;
            return TRUE;
        }
    }
    *(map+*k*2*field->nx2+*j*2+*d)=TRUE;
    return FALSE;
}
int NextPointJ(PTRData* field, int l, double val, char* map,
               int* i, int j, int* k, int* d, int* e,
               double* x3, double* x1)
{
    int m, k0, i0, k1, i1, d0, e0;
    double a, b;
    static int dd[][3]={{1, 0, 1}, {0, 1, 0}};
    static int ee[][2][3]
            ={{{0, 0, 1}, {1, 1, 0}}, {{1, 0, 0}, {0, 1, 1}}};
    static int ii[][2][3]
            ={{{1, 0, 0}, {0, 0, 1}}, {{0, 1, 0},{-1,-1,-1}}};
    static int jj[][2][3]
            ={{{0, 1, 0},{-1,-1,-1}}, {{0, 0, 1}, {1, 0, 0}}};
    d0=*d;  k0=*k;  i0=*i;
    for(m=0; m<3; m++){
        d0=dd[*d][m];
        e0=ee[*d][*e][m];
        k0=*k+ii[*d][*e][m];  i0=*i+jj[*d][*e][m];
        k1=k0+ai[d0];  i1=i0+aj[d0];
        if(*(map+i0*2*field->nx3+k0*2+d0)
           || k0<0 || field->nx3<=k0 || i0<0 || field->nx1<=i0
           || k1<0 || field->nx3<=k1 || i1<0 || field->nx1<=i1){
            continue;
        }
        *(map+i0*2*field->nx3+k0*2+d0)=TRUE;
        a=ExtractV(field, i0, j, k0, l);
        b=ExtractV(field, i1, j, k1, l);
        if((a-val)*(b-val)<=0 && b!=a){
            double x30, x31, x10, x11;
            x30=field->x3[k0];  x31=field->x3[k1];
            x10=field->x1[i0];  x11=field->x1[i1];
            *x3=x30+(val-a)/(b-a)*(x31-x30);
            *x1=x10+(val-a)/(b-a)*(x11-x10);
            *k=k0;  *i=i0;  *d=d0;  *e=e0;
            return TRUE;
        }
    }
    *(map+*i*2*field->nx3+*k*2+*d)=TRUE;
    return FALSE;
}
PTRData** ToContourDataK(PTRData* field, int l, int k,
                         double val, int* num)
{
    int i, j, d, e, nbuf, ndbuf, nd;
    char* map;
    PTRData* fld;
    PTRData** dts;
    PTRData** datas;
    double* x1;  double* x2;  double* x3=NULL;  double* v;
    EnterFunc("ToContourDataK");
    *num=0;
    if(field){
        fld=field;
    }else if(!(fld=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(l<1){
        SetError(ERROR, E_CompNumReq);
        LeaveFunc();  return NULL;
    }
    if(fld->type!=GRID){
        SetError(ERROR, E_BadDataType);
        LeaveFunc();  return NULL;
    }else if(fld->dim<2){
        SetError(ERROR, E_NoEnoughDim);
        LeaveFunc();  return NULL;
    }else if(fld->nv<l){
        SetError(ERROR, E_NoEnoughLeng);
        LeaveFunc();  return NULL;
    }
    if(fld->dim==2) k=0;
    map=(char*)malloc(2*fld->nx1*fld->nx2*sizeof(char));
    memset(map, FALSE, 2*fld->nx1*fld->nx2);
    nbuf=1024;
    x1=(double*)malloc(nbuf*sizeof(double));
    x2=(double*)malloc(nbuf*sizeof(double));
    if(fld->dim==3) x3=(double*)malloc(nbuf*sizeof(double));
    v =(double*)malloc(nbuf*sizeof(double));
    ndbuf=32;
    dts=(PTRData**)malloc(ndbuf*sizeof(PTRData*));
    i=0;  j=0;  d=0;  e=0;  nd=0;
    while(SearchStartPointK(fld, l, val, map, &i, &j, k, &d, &e)){
        int n=0;
        double a, b;
        a=ExtractV(fld, i, j, k, l);
        b=ExtractV(fld, i+ai[d], j+aj[d], k, l);
        if((a-val)*(b-val)<=0 && b!=a){
            double x10, x11, x20, x21;
            x10=fld->x1[i];  x11=fld->x1[i+ai[d]];
            x20=fld->x2[j];  x21=fld->x2[j+aj[d]];
            x1[n]=x10+(val-a)/(b-a)*(x11-x10);
            x2[n]=x20+(val-a)/(b-a)*(x21-x20);
            v[n] =val;  if(fld->dim==3) x3[n]=fld->x3[k];
            n++;
        }
        while(NextPointK(fld, l, val, map,
                         &i, &j, k, &d, &e, x1+n, x2+n)){
            v[n]=val;  if(fld->dim==3) x3[n]=fld->x3[k];
            n++;
            if(nbuf<=n){
                nbuf=2*nbuf;
                x1=realloc(x1, nbuf*sizeof(double));
                x2=realloc(x2, nbuf*sizeof(double));
                if(fld->dim==3) x3=realloc(x3, nbuf*sizeof(double));
                v =realloc(v,  nbuf*sizeof(double));
            }
        }
        if(1<n){
            dts[nd]=CreateSeqData(fld->dim, n, 1, x1, x2, x3, &v);
            nd++;
            if(ndbuf<=nd){
                ndbuf=2*ndbuf;
                dts=realloc(dts, ndbuf*sizeof(PTRData*));
            }
        }
    }
    datas=(PTRData**)malloc(nd*sizeof(PTRData*));
    memcpy(datas, dts, nd*sizeof(PTRData*));
    *num=nd;
    free(x1);  free(x2);  if(x3) free(x3);
    free(v);  free(dts);  free(map);
    LeaveFunc();  return datas;
}
PTRData** ToContourDataI(PTRData* field, int l, int i,
                         double val, int* num)
{
    int j, k, d, e, nbuf, ndbuf, nd;
    char* map;
    PTRData* fld;
    PTRData** dts;
    PTRData** datas;
    double* x1;  double* x2;  double* x3;  double* v;
    EnterFunc("ToContourDataI");
    *num=0;
    if(field){
        fld=field;
    }else if(!(fld=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(l<1){
        SetError(ERROR, E_CompNumReq);
        LeaveFunc();  return NULL;
    }
    if(fld->type!=GRID){
        SetError(ERROR, E_BadDataType);
        LeaveFunc();  return NULL;
    }else if(fld->dim<3){
        SetError(ERROR, E_NoEnoughDim);
        LeaveFunc();  return NULL;
    }else if(fld->nv<l){
        SetError(ERROR, E_NoEnoughLeng);
        LeaveFunc();  return NULL;
    }
    map=(char*)malloc(2*fld->nx2*fld->nx3*sizeof(char));
    memset(map, FALSE, 2*fld->nx2*fld->nx3);
    nbuf=1024;
    x1=(double*)malloc(nbuf*sizeof(double));
    x2=(double*)malloc(nbuf*sizeof(double));
    x3=(double*)malloc(nbuf*sizeof(double));
    v =(double*)malloc(nbuf*sizeof(double));
    ndbuf=32;
    dts=(PTRData**)malloc(ndbuf*sizeof(PTRData*));
    j=0;  k=0;  d=0;  e=0;  nd=0;
    while(SearchStartPointI(fld, l, val, map, i, &j, &k, &d, &e)){
        int n=0;
        double a, b;
        a=ExtractV(fld, i, j, k, l);
        b=ExtractV(fld, i, j+ai[d], k+aj[d], l);
        if((a-val)*(b-val)<=0 && b!=a){
            double x20, x21, x30, x31;
            x20=fld->x2[j];  x21=fld->x2[j+ai[d]];
            x30=fld->x3[k];  x31=fld->x3[k+aj[d]];
            x2[n]=x20+(val-a)/(b-a)*(x21-x20);
            x3[n]=x30+(val-a)/(b-a)*(x31-x30);
            v[n] =val;  x1[n]=fld->x1[i];
            n++;
        }
        while(NextPointI(fld, l, val, map,
                         i, &j, &k, &d, &e, x2+n, x3+n)){
            v[n]=val;  x1[n]=fld->x1[i];
            n++;
            if(nbuf<=n){
                nbuf=2*nbuf;
                x1=realloc(x1, nbuf*sizeof(double));
                x2=realloc(x2, nbuf*sizeof(double));
                x3=realloc(x3, nbuf*sizeof(double));
                v =realloc(v,  nbuf*sizeof(double));
            }
        }
        if(1<n){
            dts[nd]=CreateSeqData(fld->dim, n, 1, x1, x2, x3, &v);
            nd++;
            if(ndbuf<=nd){
                ndbuf=2*ndbuf;
                dts=realloc(dts, ndbuf*sizeof(PTRData*));
            }
        }
    }
    datas=(PTRData**)malloc(nd*sizeof(PTRData*));
    memcpy(datas, dts, nd*sizeof(PTRData*));
    *num=nd;
    free(x1);  free(x2);  free(x3);
    free(v);  free(dts);  free(map);
    LeaveFunc();  return datas;
}
PTRData** ToContourDataJ(PTRData* field, int l, int j,
                         double val, int* num)
{
    int k, i, d, e, nbuf, ndbuf, nd;
    char* map;
    PTRData* fld;
    PTRData** dts;
    PTRData** datas;
    double* x1;  double* x2;  double* x3;  double* v;
    EnterFunc("ToContourDataJ");
    *num=0;
    if(field){
        fld=field;
    }else if(!(fld=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(l<1){
        SetError(ERROR, E_CompNumReq);
        LeaveFunc();  return NULL;
    }
    if(fld->type!=GRID){
        SetError(ERROR, E_BadDataType);
        LeaveFunc();  return NULL;
    }else if(fld->dim<3){
        SetError(ERROR, E_NoEnoughDim);
        LeaveFunc();  return NULL;
    }else if(fld->nv<l){
        SetError(ERROR, E_NoEnoughLeng);
        LeaveFunc();  return NULL;
    }
    map=(char*)malloc(2*fld->nx3*fld->nx1*sizeof(char));
    memset(map, FALSE, 2*fld->nx3*fld->nx1);
    nbuf=1024;
    x1=(double*)malloc(nbuf*sizeof(double));
    x2=(double*)malloc(nbuf*sizeof(double));
    x3=(double*)malloc(nbuf*sizeof(double));
    v =(double*)malloc(nbuf*sizeof(double));
    ndbuf=32;
    dts=(PTRData**)malloc(ndbuf*sizeof(PTRData*));
    k=0;  i=0;  d=0;  e=0;  nd=0;
    while(SearchStartPointJ(fld, l, val, map, &i, j, &k, &d, &e)){
        int n=0;
        double a, b;
        a=ExtractV(fld, i, j, k, l);
        b=ExtractV(fld, i+aj[d], j, k+ai[d], l);
        if((a-val)*(b-val)<=0 && b!=a){
            double x30, x31, x10, x11;
            x30=fld->x3[k];  x31=fld->x3[k+ai[d]];
            x10=fld->x1[i];  x11=fld->x1[i+aj[d]];
            x3[n]=x30+(val-a)/(b-a)*(x31-x30);
            x1[n]=x10+(val-a)/(b-a)*(x11-x10);
            v[n] =val;  x2[n]=fld->x2[j];
            n++;
        }
        while(NextPointJ(fld, l, val, map,
                         &i, j, &k, &d, &e, x3+n, x1+n)){
            v[n]=val;  x2[n]=fld->x2[j];
            n++;
            if(nbuf<=n){
                nbuf=2*nbuf;
                x1=realloc(x1, nbuf*sizeof(double));
                x2=realloc(x2, nbuf*sizeof(double));
                x3=realloc(x3, nbuf*sizeof(double));
                v =realloc(v,  nbuf*sizeof(double));
            }
        }
        if(1<n){
            dts[nd]=CreateSeqData(fld->dim, n, 1, x1, x2, x3, &v);
            nd++;
            if(ndbuf<=nd){
                ndbuf=2*ndbuf;
                dts=realloc(dts, ndbuf*sizeof(PTRData*));
            }
        }
    }
    datas=(PTRData**)malloc(nd*sizeof(PTRData*));
    memcpy(datas, dts, nd*sizeof(PTRData*));
    *num=nd;
    free(x1);  free(x2);  free(x3);
    free(v);  free(dts);  free(map);
    LeaveFunc();  return datas;
}

PTRData** ToContourData(PTRData* field, int l,
                        char dir, int nn, double val, int* num)
{
    PTRData* fld;
    EnterFunc("ToContourData");
    *num=0;
    if(field){
        fld=field;
    }else if(!(fld=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(fld->dim==2 || dir=='k'){
        return ToContourDataK(fld, l, nn, val, num);
    }else if(dir=='i'){
        return ToContourDataI(fld, l, nn, val, num);
    }else if(dir=='j'){
        return ToContourDataJ(fld, l, nn, val, num);
    }else{
        SetError(ERROR, E_BadValue);
        LeaveFunc();  return NULL;
    }
}

/* stream line */
PTRData* ToStLineE_FB(PTRData* fld, int l1, int l2, int l3,
                      double sx1, double sx2, double sx3, int num)
{
    int nmax, nv=2;
    int vi[3];
    int ista, iend, it, i, nt;
    double vt[3];
    double stp, ds, lmax;
    double min, max;
    double x1size, x2size;  double x3size=0.0;
    double px3, v1, v2, v3, vl;
    double*  x1;  double*  x2;  double*  x3=NULL;
    double* tx1;  double* tx2;  double* tx3=NULL;
    PTRData* dp;
    vi[0]=l1;  vi[1]=l2;
    nmax=StreamLineNum();
    ExtractX1Range(fld, &min, &max);  x1size=max-min;
    ExtractX2Range(fld, &min, &max);  x2size=max-min;
    if(2<fld->dim){
        if(0<l3){
            nv=3;  vi[2]=l3;
        }
        ExtractX3Range(fld, &min, &max);  x3size=max-min;
    }
    lmax=StreamLineLength()
            *sqrt(x1size*x1size+x2size*x2size+x3size*x3size);
    ds=lmax/nmax;
    tx1=(double*)malloc(2*nmax*sizeof(double));
    tx2=(double*)malloc(2*nmax*sizeof(double));
    if(2<fld->dim) tx3=(double*)malloc(2*nmax*sizeof(double));
    tx1[nmax]=sx1;  tx2[nmax]=sx2;  if(2<fld->dim) tx3[nmax]=sx3;
    for(iend=nmax+1; iend<2*nmax; iend++){
        px3=0.0;  v3=0.0;
        if(2<fld->dim) px3=tx3[iend-1];
        if(!(InterpolateVector(fld, tx1[iend-1], tx2[iend-1],
                               px3, nv, vi, vt))) break;
        v1=vt[0];  v2=vt[1];  if(nv==3) v3=vt[2];
        vl=sqrt(v1*v1+v2*v2+v3*v3);
        tx1[iend]=tx1[iend-1]+v1/vl*ds;
        tx2[iend]=tx2[iend-1]+v2/vl*ds;
        if(2<fld->dim) tx3[iend]=tx3[iend-1]+v3/vl*ds;
    }
    iend--;
    for(ista=nmax-1; 0<=ista; ista--){
        px3=0.0;  v3=0.0;
        if(2<fld->dim) px3=tx3[ista+1];
        if(!(InterpolateVector(fld, tx1[ista+1], tx2[ista+1],
                               px3, nv, vi, vt))) break;
        v1=vt[0];  v2=vt[1];  if(nv==3) v3=vt[2];
        vl=sqrt(v1*v1+v2*v2+v3*v3);
        tx1[ista]=tx1[ista+1]-v1/vl*ds;
        tx2[ista]=tx2[ista+1]-v2/vl*ds;
        if(2<fld->dim) tx3[ista]=tx3[ista+1]-v3/vl*ds;
    }
    ista++;
    nt=iend-ista+1;
    if(nt<num) num=nt;
    x1=(double*)malloc(num*sizeof(double));
    x2=(double*)malloc(num*sizeof(double));
    if(2<fld->dim) x3=(double*)malloc(num*sizeof(double));
    stp=(double)(nt-1)/(num-1);
    for(i=0; i<num; i++){
        it=ista+floor(stp*i+0.5);
        x1[i]=tx1[it];  x2[i]=tx2[it];
        if(2<fld->dim) x3[i]=tx3[it];
    }
    i=floor((nmax-ista)/stp+0.5);
    x1[i]=tx1[nmax];  x2[i]=tx2[nmax];
    if(2<fld->dim) x3[i]=tx3[nmax];
    dp=CreateSeqData(fld->dim, num, 0, x1, x2, x3, NULL);
    free( x1);  free( x2);  if( x3) free( x3);
    free(tx1);  free(tx2);  if(tx3) free(tx3);
    return dp;
}
PTRData* ToStreamLine(PTRData* field, int l1, int l2, int l3,
                      double x1, double x2, double x3, int num)
{
    PTRData* fld;
    EnterFunc("ToStreamLine");
    if(field){
        fld=field;
    }else if(!(fld=CurrentData())){
        SetError(ERROR, E_NoData);
        LeaveFunc();  return NULL;
    }
    if(l1<1 || l2<1){
        SetError(ERROR, E_CompNumReq);
        LeaveFunc();  return NULL;
    }
    if(fld->type != GRID){
        SetError(ERROR, E_BadDataType);
        LeaveFunc();  return NULL;
    }else if(fld->dim<2){
        SetError(ERROR, E_NoEnoughDim);
        LeaveFunc();  return NULL;
    }else if(fld->nv<l1 || fld->nv<l2 || fld->nv<l3){
        SetError(ERROR, E_NoEnoughLeng);
        LeaveFunc();  return NULL;
    }
    LeaveFunc();
    if(stlalg==RUNGE){
        if(stldir==BACKWARD){
            return ToStLineE_FB(fld, l1, l2, l3, x1, x2, x3, num);
        }else if(stldir==BOTH){
            return ToStLineE_FB(fld, l1, l2, l3, x1, x2, x3, num);
        }else{
            return ToStLineE_FB(fld, l1, l2, l3, x1, x2, x3, num);
        }
    }else{
        if(stldir==BACKWARD){
            return ToStLineE_FB(fld, l1, l2, l3, x1, x2, x3, num);
        }else if(stldir==BOTH){
            return ToStLineE_FB(fld, l1, l2, l3, x1, x2, x3, num);
        }else{
            return ToStLineE_FB(fld, l1, l2, l3, x1, x2, x3, num);
        }
    }
}
void StreamLineMode(int alg, int dir){
    stlalg=alg;  stldir=dir;
}
