#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pgw.h>
#include "ptr.h"



int OverWrite(char* fname)
{
    return TRUE;
}





/* set scales */

PTRScale* CreatePTRScale(int ax, char* form)
{
    PTRScale* gp;
    gp=(PTRScale*)malloc(sizeof(PTRScale));
    gp->type=SCALE;
    gp->autoset=TRUE;
    gp->ax=ax;
    gp->nx=0;
    gp->scl=NULL;
    gp->x1=NULL;
    gp->x2=NULL;
    gp->x3=NULL;
    gp->sc_dist =0.5*SCL_DIST;
    gp->scl_dist=SCL_DIST;
    gp->lb_dist =LB_DIST;
    if(form){
        strcpy(gp->form, form);
    }else{
        strcpy(gp->form, SCL_FORM);
    }
    gp->pgp=GetPGP();
    gp->pgp.pt=0.9*(gp->pgp.pt);
    gp->pgp.pgcl=DefaultForeground();
    return gp;
}


void ScaleStep(double range, double* scstp, double* sclstp)
{
    double eps=1.0e-8;
    EnterFunc("ScaleStep");
    if(1.0-eps<=range && range<1.5){
        *scstp=0.05;  *sclstp=0.25;
    }else if(1.5<=range && range<3.0){
        *scstp=0.1;   *sclstp=0.5;
    }else if(3.0<=range && range<5.0){
        *scstp=0.2;   *sclstp=1.0;
    }else if(5.0<=range && range<6.5){
        *scstp=0.5;   *sclstp=1.0;
    }else if(6.5<=range && range<10.0){
        *scstp=0.5;   *sclstp=2.0;
    }else{
        char msg[MAX_STR];
        *scstp=0.5;   *sclstp=2.0;
        sprintf(msg, "Can't set scale steps.\nrange=%lf\n", range);
        SetError(INNER, msg);
    }
    LeaveFunc();
}


int AutoScaleLabel(double min, double max,
                   char* form, ScaleLabel** scl)
{
    int n=0;
    int i;
    double order, size, range, val, scstp, sclstp;
    double eps=1.0e-8;
    ScaleLabel scl1[32];
    size=max-min;
    order=floor(log10(size));
    range=size*pow(10.0, -order);
    ScaleStep(range, &scstp, &sclstp);
    scstp =scstp*pow(10.0, order);
    sclstp=sclstp*pow(10.0, order);
    for(val=scstp*ceil(min/scstp); val<=max+eps; val+=scstp){
        scl1[n].x=val;
        if(fabs((val/sclstp)-floor(val/sclstp+0.5))<eps){
            char s[256];
            if(fabs(val/sclstp)<eps) val=0.0;
            sprintf(s, form, val);
            for(i=strlen(s)-1; s[i]==' '; i--) s[i]='\0';
            sprintf(scl1[n].lb, "%s", s);
        }else{
            scl1[n].lb[0]='\0';
        }
        n++;
    }
    *scl=(ScaleLabel*)malloc(n*sizeof(ScaleLabel));
    for(i=0; i<n; i++) (*scl)[i]=scl1[i];
    return n;
}

void RemoveSclFrame(PTRWindow* wp)
{
    if(wp->frame==NULL) return;
    RemoveGraph(wp, (PTRGraph*)(wp->frame));
    wp->frame=NULL;
}
void SetSclFrame(PTRWindow* wp, PGP pgp)
{
    PTRSclFrame* gp;
    RemoveSclFrame(wp);
    gp=(PTRSclFrame*)malloc(sizeof(PTRSclFrame));
    gp->type=SCLFRAME;
    gp->pgp=pgp;
    gp->pgp.pgcl=DefaultForeground();
    wp->frame=gp;
    AddPTRGraph(wp, (PTRGraph*)gp);
}
void SetAxScale(PTRWindow* wp, int ax, int num, ScaleLabel scl[])
{
    int i;
    PTRScale* gp;
    gp=wp->scale[ax];
    if(gp==NULL){
        gp=CreatePTRScale(ax, NULL);
    }
    gp->autoset=FALSE;
    gp->num=num;
    if(wp->frame==NULL) SetSclFrame(wp, GetPGP());
    if(gp->scl){
        gp->scl=(ScaleLabel*)
                realloc(gp->scl, gp->num*sizeof(ScaleLabel));
    }else{
        gp->scl=(ScaleLabel*)malloc(gp->num*sizeof(ScaleLabel));
    }
    for(i=0; i<gp->num; i++) gp->scl[i]=scl[i];
    if(wp->scale[ax]){
        UpdatePTRGraph(wp, (PTRGraph*)gp);
    }else{
        wp->scale[ax]=gp;
        AddPTRGraph(wp, (PTRGraph*)gp);
    }
}
void AddAxScale(PTRWindow* wp, int ax, double x, char* label)
{
    PTRScale* gp;
    gp=wp->scale[ax];
    if(gp==NULL){
        gp=CreatePTRScale(ax, NULL);
    }
    gp->autoset=FALSE;
    if(wp->frame==NULL) SetSclFrame(wp, GetPGP());
    gp->num++;
    if(gp->scl){
        gp->scl=(ScaleLabel*)
                realloc(gp->scl, gp->num*sizeof(ScaleLabel));
    }else{
        gp->scl=(ScaleLabel*)malloc(gp->num*sizeof(ScaleLabel));
    }
    gp->scl[gp->num-1].x=x;
    gp->scl[gp->num-1].lb[0]='\0';
    if(label) strcpy(gp->scl[gp->num-1].lb, label);
    if(wp->scale[ax]){
        UpdatePTRGraph(wp, (PTRGraph*)gp);
    }else{
        wp->scale[ax]=gp;
        AddPTRGraph(wp, (PTRGraph*)gp);
    }
}
void AutoAxScale(PTRWindow* wp, int ax, char* form)
{
    PTRScale* gp;
    gp=wp->scale[ax];
    if(gp==NULL){
        gp=CreatePTRScale(ax, form);
    }
    gp->autoset=TRUE;
    if(form) strcpy(gp->form, form);
    if(wp->frame==NULL) SetSclFrame(wp, GetPGP());
    if(gp->scl) free(gp->scl);
    gp->num=AutoScaleLabel(wp->min[ax], wp->max[ax],
                           gp->form, &(gp->scl));
    if(wp->scale[ax]){
        UpdatePTRGraph(wp, (PTRGraph*)gp);
    }else{
        wp->scale[ax]=gp;
        AddPTRGraph(wp, (PTRGraph*)gp);
    }
}
void SetX1Scale(PTRWindow* wp, int num, ScaleLabel scl[]){
    SetAxScale(wp, AX_X1, num, scl);
}
void AddX1Scale(PTRWindow* wp, double x, char* label){
    AddAxScale(wp, AX_X1, x, label);
}
void AutoX1Scale(PTRWindow* wp, char* form){
    AutoAxScale(wp, AX_X1, form);
}
void RemoveScale(PTRWindow* wp)
{
    int i;
    for(i=1; i<5; i++){
        if(wp->scale[i]){
            RemoveGraph(wp, (PTRGraph*)(wp->scale[i]));
            wp->scale[i]=NULL;
        }
    }
    if(wp->frame){
        RemoveSclFrame(wp);
    }
}
void SetScale(PTRWindow* wp, char* form)
{
    int i;
    for(i=1; i<5; i++) AutoAxScale(wp, i, form);
}
void UpdateScales(PTRWindow* wp)
{
    int i;
    for(i=1; i<5; i++)
            UpdatePTRGraph(wp, (PTRGraph*)(wp->scale[i]));
    UpdatePTRGraph(wp, (PTRGraph*)(wp->frame));
}
void UpdateLabels(PTRWindow* wp)
{
    int i;
    for(i=0; i<5; i++)
            UpdatePTRGraph(wp, (PTRGraph*)(wp->label[i]));
}

/* set ratio of each axis */
void ExactRatio(PTRWindow* wp, int val)
{
    if(val){
        double x2_x1, x3_x1;
        wp->exact=TRUE;
        x2_x1=wp->size[wp->ax2]/wp->size[wp->ax1];
        SetPGWindowX2_X1(wp->pgw, x2_x1);
        if(wp->ax3){
            x3_x1=wp->size[wp->ax3]/wp->size[wp->ax1];
            SetPGWindowX3_X1(wp->pgw, x3_x1);
        }
    }else{
        wp->exact=FALSE;
        SetPGWindowX2_X1(wp->pgw, 0.0);
        SetPGWindowX3_X1(wp->pgw, 0.0);
    }
}
void SetY_X(PTRWindow* wp, double rat){
    SetPGWindowX2_X1(wp->pgw, rat);
    wp->exact=FALSE;
}
void SetZ_X(PTRWindow* wp, double rat){
    SetPGWindowX3_X1(wp->pgw, rat);
    wp->exact=FALSE;
}

/* set limits */

void SetPGWAxLimit(PTRWindow* wp, int ax, double min, double max)
{
    if(wp->ax1==ax) SetPGWindowX1Limit(wp->pgw, min, max);
    if(wp->ax2==ax) SetPGWindowX2Limit(wp->pgw, min, max);
    if(wp->ax3==ax) SetPGWindowX3Limit(wp->pgw, min, max);
    UpdateScales(wp);
    UpdateLabels(wp);
}
void SetAxLimit(PTRWindow* wp, int ax, double min, double max)
{
    EnterFunc("SetAxLimit");
    if(max<=min){
        SetError(ERROR, E_BadValue);
        LeaveFunc();  return;
    }
    wp->min[ax]=min;  wp->max[ax]=max;
    wp->size[ax]=max-min;
    SetPGWAxLimit(wp, ax, min, max);
    if(wp->exact) ExactRatio(wp, TRUE);
    if(wp->scale[ax] && wp->scale[ax]->autoset)
            AutoAxScale(wp, ax, NULL);
    LeaveFunc();
}

double X1Min(PTRWindow* wp){
    return wp->min[AX_X1];
}
double X1Max(PTRWindow* wp){
    return wp->max[AX_X1];
}
void SetX1Limit(PTRWindow* wp, double min, double max){
    EnterFunc("SetX1Limit");
    SetAxLimit(wp, AX_X1, min, max);
    LeaveFunc();
}
void SetX1Min(PTRWindow* wp, double min){
    EnterFunc("SetX1Min");
    SetAxLimit(wp, AX_X1, min, wp->max[AX_X1]);
    LeaveFunc();
}
void SetX1Max(PTRWindow* wp, double max){
    EnterFunc("SetX1Max");
    SetAxLimit(wp, AX_X1, wp->min[AX_X1], max);
    LeaveFunc();
}
double X2Min(PTRWindow* wp){
    return wp->min[AX_X2];
}
double X2Max(PTRWindow* wp){
    return wp->max[AX_X2];
}
void SetX2Limit(PTRWindow* wp, double min, double max){
    EnterFunc("SetX2Limit");
    SetAxLimit(wp, AX_X2, min, max);
    LeaveFunc();
}
void SetX2Min(PTRWindow* wp, double min){
    EnterFunc("SetX2Min");
    SetAxLimit(wp, AX_X2, min, wp->max[AX_X2]);
    LeaveFunc();
}
void SetX2Max(PTRWindow* wp, double max){
    EnterFunc("SetX2Max");
    SetAxLimit(wp, AX_X2, wp->min[AX_X2], max);
    LeaveFunc();
}
double X3Min(PTRWindow* wp){
    return wp->min[AX_X3];
}
double X3Max(PTRWindow* wp){
    return wp->max[AX_X3];
}
void SetX3Limit(PTRWindow* wp, double min, double max){
    EnterFunc("SetX3Limit");
    SetAxLimit(wp, AX_X3, min, max);
    LeaveFunc();
}
void SetX3Min(PTRWindow* wp, double min){
    EnterFunc("SetX3Min");
    SetAxLimit(wp, AX_X3, min, wp->max[AX_X3]);
    LeaveFunc();
}
void SetX3Max(PTRWindow* wp, double max){
    EnterFunc("SetX3Max");
    SetAxLimit(wp, AX_X3, wp->min[AX_X3], max);
    LeaveFunc();
}
double VMin(PTRWindow* wp){
    return wp->min[AX_V];
}
double VMax(PTRWindow* wp){
    return wp->max[AX_V];
}
void SetVLimit(PTRWindow* wp, double min, double max){
    EnterFunc("SetVLimit");
    SetAxLimit(wp, AX_V, min, max);
    LeaveFunc();
}
void SetVMin(PTRWindow* wp, double min){
    EnterFunc("SetVMin");
    SetAxLimit(wp, AX_V, min, wp->max[AX_V]);
    LeaveFunc();
}
void SetVMax(PTRWindow* wp, double max){
    EnterFunc("SetVMax");
    SetAxLimit(wp, AX_V, wp->min[AX_V], max);
    LeaveFunc();
}

double XMin(PTRWindow* wp){
    return wp->min[wp->ax1];
}
double XMax(PTRWindow* wp){
    return wp->max[wp->ax1];
}
void SetXLimit(PTRWindow* wp, double min, double max){
    EnterFunc("SetXLimit");
    SetAxLimit(wp, wp->ax1, min, max);
    LeaveFunc();
}
void SetXMin(PTRWindow* wp, double min){
    EnterFunc("SetXMin");
    SetAxLimit(wp, wp->ax1, min, wp->max[wp->ax1]);
    LeaveFunc();
}
void SetXMax(PTRWindow* wp, double max){
    EnterFunc("SetXMax");
    SetAxLimit(wp, wp->ax1, wp->min[wp->ax1], max);
    LeaveFunc();
}
double YMin(PTRWindow* wp){
    return wp->min[wp->ax2];
}
double YMax(PTRWindow* wp){
    return wp->max[wp->ax2];
}
void SetYLimit(PTRWindow* wp, double min, double max){
    EnterFunc("SetYLimit");
    SetAxLimit(wp, wp->ax2, min, max);
    LeaveFunc();
}
void SetYMin(PTRWindow* wp, double min){
    EnterFunc("SetYMin");
    SetAxLimit(wp, wp->ax2, min, wp->max[wp->ax2]);
    LeaveFunc();
}
void SetYMax(PTRWindow* wp, double max){
    EnterFunc("SetYMax");
    SetAxLimit(wp, wp->ax2, wp->min[wp->ax2], max);
    LeaveFunc();
}
double ZMin(PTRWindow* wp){
    return wp->min[wp->ax3];
}
double ZMax(PTRWindow* wp){
    return wp->max[wp->ax3];
}
void SetZLimit(PTRWindow* wp, double min, double max){
    EnterFunc("SetZLimit");
    SetAxLimit(wp, wp->ax3, min, max);
    LeaveFunc();
}
void SetZMin(PTRWindow* wp, double min){
    EnterFunc("SetZMin");
    SetAxLimit(wp, wp->ax3, min, wp->max[wp->ax3]);
    LeaveFunc();
}
void SetZMax(PTRWindow* wp, double max){
    EnterFunc("SetZMax");
    SetAxLimit(wp, wp->ax3, wp->min[wp->ax3], max);
    LeaveFunc();
}


/* change coordinate */

void SetCoord(PTRWindow* wp, int coord)
{
    EnterFunc("SetCoord");
    switch(coord){
      case CAR:
        SetPGWindowCoord(wp->pgw, CAR);
        break;
      case CYL:
        SetPGWindowCoord(wp->pgw, CYL);
        break;
      default:
        SetError(ERROR, E_BadValue);
        LeaveFunc();  return;
    }
    wp->coord=coord;
    UpdateScales(wp);
    LeaveFunc();
}
void SetX1Axis(PTRWindow* wp, int ax)
{
    if(ax<1 || 4<ax) return;
    wp->ax1=ax;
    SetPGWindowX1Limit(wp->pgw, wp->min[ax], wp->max[ax]);
    UpdatePTRGraphs(wp);
    if(wp->exact) ExactRatio(wp, TRUE);
}
void SetX2Axis(PTRWindow* wp, int ax)
{
    if(ax<1 || 4<ax) return;
    wp->ax2=ax;
    SetPGWindowX2Limit(wp->pgw, wp->min[ax], wp->max[ax]);
    UpdatePTRGraphs(wp);
    if(wp->exact) ExactRatio(wp, TRUE);
}
void SetX3Axis(PTRWindow* wp, int ax)
{
    if(ax<0 || 4<ax) return;
    if(wp->ax3 && ax==AX_NL){
        SetPGWindowDimension(wp->pgw, DIM2);
    }else if(wp->ax3==AX_NL && ax){
        SetPGWindowDimension(wp->pgw, DIM3);
        SetPGWindowX3Limit(wp->pgw, wp->min[ax], wp->max[ax]);
    }
    wp->ax3=ax;
    UpdatePTRGraphs(wp);
    if(wp->exact) ExactRatio(wp, TRUE);
}
void SetAxis(PTRWindow* wp, int ax1, int ax2, int ax3)
{
    if(ax1<1 || 4<ax1) return;
    if(ax2<1 || 4<ax2) return;
    if(ax3<0 || 4<ax3) return;
    SetPGWindowX1Limit(wp->pgw, wp->min[ax1], wp->max[ax1]);
    SetPGWindowX2Limit(wp->pgw, wp->min[ax2], wp->max[ax2]);
    if(wp->ax3 && ax3==AX_NL){
        SetPGWindowDimension(wp->pgw, DIM2);
    }else if(wp->ax3==AX_NL && ax3){
        SetPGWindowDimension(wp->pgw, DIM3);
        SetPGWindowX3Limit(wp->pgw, wp->min[ax3], wp->max[ax3]);
    }
    wp->ax1=ax1;  wp->ax2=ax2;  wp->ax3=ax3;
    UpdatePTRGraphs(wp);
    if(wp->exact) ExactRatio(wp, TRUE);
}

/* change angle */
void SetPhi(PTRWindow* wp, double ang){
    SetPhiAngle(wp->pgw, ang);
}
void SetTheta(PTRWindow* wp, double ang){
    SetThetaAngle(wp->pgw, ang);
}

/* change magnifying const */
void SetMagnify(PTRWindow* wp, double mag){
    SetMag(wp->pgw, mag);
}
/* change origin */
void SetOriginX(PTRWindow* wp, double ox){
    SetOrgx(wp->pgw, ox);
}
void SetOriginY(PTRWindow* wp, double oy){
    SetOrgy(wp->pgw, oy);
}
