#include "../config.h"
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "pgw.h"

#ifdef USE_LIB_XPM
#include <X11/xpm.h>
#endif /* USE_LIB_XPM */
#if defined(USE_LIB_UNGIF) || defined(USE_LIB_GIF)
#include <gif_lib.h>
#endif /* USE_LIB_UNGIF || USE_LIB_GIF */

Display* display;
Colormap cm;
int nowindow=FALSE;

PGWindow pgroot;


PGWindow* XWindowToPGWindow(PGWindow* pgwindow, Window w)
{
    PGWindowList* pgwl;
    pgwl=pgwindow->pgwl.next;
    while(pgwl){
        PGWindow* ret=NULL;
        if(pgwl->pgw->w==w){
            return pgwl->pgw;
        }else if((ret=XWindowToPGWindow(pgwl->pgw, w))){
            return ret;
        }
        pgwl=pgwl->next;
    }
    return NULL;
}

void AddPGWindow(PGWindow* parent, PGWindow* pgw)
{
    PGWindowList* pgwl;
    pgwl=(PGWindowList*)malloc(sizeof(PGWindowList));
    pgwl->pgw=pgw;
    pgwl->next=parent->pgwl.next;
    parent->pgwl.next=pgwl;
}



/* get color */


int PGGetColor(int r, int g, int b, PGColor* pgcl)
{
    if(pgcl==NULL) return FALSE;
    if(r<0 || 255<r) return FALSE;
    if(g<0 || 255<g) return FALSE;
    if(b<0 || 255<b) return FALSE;
    if(nowindow){
        pgcl->r=r;  pgcl->g=g;  pgcl->b=b;
        return TRUE;
    }else{
        XColor xcl;
        xcl.red  =(unsigned short)(r/255.0*65535.0+0.5);
        xcl.green=(unsigned short)(g/255.0*65535.0+0.5);
        xcl.blue =(unsigned short)(b/255.0*65535.0+0.5);
        if(XAllocColor(display, cm, &xcl)){
            pgcl->xcolor=xcl;
            pgcl->r=r;  pgcl->g=g;  pgcl->b=b;
            return TRUE;
        }else{
            return FALSE;
        }
    }
}
int PGGetColors(int n, int r[], int g[], int b[], PGColor pgcl[])
{
    int i;
    if(pgcl==NULL || n<1) return FALSE;
    for(i=0; i<n; i++){
        if(r[i]<0 || 255<r[i]) return FALSE;
        if(g[i]<0 || 255<g[i]) return FALSE;
        if(b[i]<0 || 255<b[i]) return FALSE;
    }
    if(nowindow){
        for(i=0; i<n; i++){
            pgcl[i].r=r[i];  pgcl[i].g=g[i];  pgcl[i].b=b[i];
        }
        return TRUE;
    }else{
        XColor xcl;
        for(i=0; i<n; i++){
            xcl.red  =(unsigned short)(r[i]/255.0*65535.0+0.5);
            xcl.green=(unsigned short)(g[i]/255.0*65535.0+0.5);
            xcl.blue =(unsigned short)(b[i]/255.0*65535.0+0.5);
/*            xcl.red  =(unsigned short)(r[i]<<8);
            xcl.green=(unsigned short)(g[i]<<8);
            xcl.blue =(unsigned short)(b[i]<<8);*/
            if(XAllocColor(display, cm, &xcl)){
                pgcl[i].xcolor=xcl;
                pgcl[i].r=r[i];  pgcl[i].g=g[i];  pgcl[i].b=b[i];
            }else{
                unsigned long* pixs;
                pixs=(unsigned long*)malloc(i*sizeof(unsigned long));
                n=i;
                for(i=0; i<n; i++) pixs[i]=pgcl[i].xcolor.pixel;
                XFreeColors(display, cm, pixs, n, 0);
                free(pixs);
                return FALSE;
            }
        }
        return TRUE;
    }
}

int PGGetNamedColor(char* name, PGColor* pgcl)
{
    if(pgcl==NULL) return FALSE;
    if(nowindow){
        int r, g, b;
        char* nm;
        FILE* fp;
        if(!(fp=fopen(RGB_TXT, "r"))) return FALSE;
        if(strcmp(name, "default")==0){
            pgcl->r=0;  pgcl->g=0;  pgcl->b=0;
            return TRUE;
        }
        while(fscanf(fp, "%d %d %d %s", &r, &g, &b, nm)==4){
            if(strcmp(nm, name)==0){
                pgcl->r=r;  pgcl->g=g;  pgcl->b=b;
                fclose(fp);
                return TRUE;
            }
        }
        fclose(fp);
        return FALSE;
    }else{
        XColor xcl, excl;
        if(strcmp(name, "default")==0
           && XAllocNamedColor(display, cm, "white", &xcl, &excl)){
            pgcl->xcolor = xcl;
            pgcl->r=0;  pgcl->g=0;  pgcl->b=0;
            return TRUE;
        }else if(XAllocNamedColor(display, cm, name, &xcl, &excl)){
            pgcl->xcolor=xcl;
            pgcl->r=(int)(excl.red  /65535.0*255.0+0.5);
            pgcl->g=(int)(excl.green/65535.0*255.0+0.5);
            pgcl->b=(int)(excl.blue /65535.0*255.0+0.5);
            return TRUE;
        }else{
            return FALSE;
        }
    }
}




int PGWSetUp(char* serv)
{
    nowindow=FALSE;
    pgroot.pgwl.next=NULL;
    pgroot.w=(Window)NULL;
    display=XOpenDisplay(serv);
    if(display){
        cm=XDefaultColormap(display, XDefaultScreen(display));
        InitializePGP();
        return TRUE;
    }else{
        return FALSE;
    }
}

int PGWPSSetUp()
{
    nowindow=TRUE;
    pgroot.pgwl.next=NULL;
    pgroot.w=(Window)NULL;
    return TRUE;
}

void ReconfigurePGWindow(PGWindow* pgw){
    MoveResizePGWindow(pgw, pgw->x, pgw->y,
                       pgw->width, pgw->height);
}
void MovePGWindow(PGWindow* pgw, double x, double y)
{
    int gmx, gmy;
    if(IsPGFrame(pgw)) return;
    pgw->x=x;  pgw->y=y;
    if(nowindow) return;
    gmx=(int)(x*pgw->parent->wdx);
    gmy=(int)(y*pgw->parent->wdy);
    XMoveWindow(display, pgw->w, gmx, gmy);
}
void ResizePGWindow(PGWindow* pgw, double width, double height)
{
    unsigned int wdx, wdy;
    PGWindowList* pgwl;
    if(IsPGFrame(pgw)) return;
    pgw->width=width;  pgw->height=height;
    if(nowindow) return;
    wdx=(unsigned int)(width*pgw->parent->wdx);
    wdy=(unsigned int)(height*pgw->parent->wdy);
    XResizeWindow(display, pgw->w, wdx, wdy);
    pgw->wdx=wdx;  pgw->wdy=wdy;
    if(pgw->p) XFreePixmap(display, pgw->p);
    pgw->p=XCreatePixmap(display, pgw->w, wdx, wdy,
                         XDefaultDepth(display, 0));
    ShowPGWindow(pgw);
    pgwl=pgw->pgwl.next;
    while(pgwl){
        ReconfigurePGWindow(pgwl->pgw);
        pgwl=pgwl->next;
    }
}
void MoveResizePGWindow(PGWindow* pgw,
                        double x, double y,
                        double width, double height)
{
    int gmx, gmy;
    unsigned int wdx, wdy;
    PGWindowList* pgwl;
    if(IsPGFrame(pgw)) return;
    pgw->x=x;  pgw->y=y;
    pgw->width=width;  pgw->height=height;
    if(nowindow) return;
    gmx=(int)(x*pgw->parent->wdx);
    gmy=(int)(y*pgw->parent->wdy);
    wdx=(unsigned int)(width*pgw->parent->wdx);
    wdy=(unsigned int)(height*pgw->parent->wdy);
    XMoveResizeWindow(display, pgw->w, gmx, gmy, wdx, wdy);
    pgw->wdx=wdx;  pgw->wdy=wdy;
    if(pgw->p) XFreePixmap(display, pgw->p);
    pgw->p=XCreatePixmap(display, pgw->w, wdx, wdy,
                         XDefaultDepth(display, 0));
    ShowPGWindow(pgw);
    pgwl=pgw->pgwl.next;
    while(pgwl){
        ReconfigurePGWindow(pgwl->pgw);
        pgwl=pgwl->next;
    }
}
void ResizePGFrame(PGWindow* pgw,
                   unsigned int wdx, unsigned int wdy)
{
    PGWindowList* pgwl;
    if((!IsPGFrame(pgw)) || nowindow) return;
    pgw->wdx=wdx;  pgw->wdy=wdy;
    XResizeWindow(display, pgw->w, wdx, wdy);
    if(pgw->p) XFreePixmap(display, pgw->p);
    pgw->p=XCreatePixmap(display, pgw->w, wdx, wdy,
                         XDefaultDepth(display, 0));
    ShowPGWindow(pgw);
    pgwl=pgw->pgwl.next;
    while(pgwl){
        ReconfigurePGWindow(pgwl->pgw);
        pgwl=pgwl->next;
    }
}

void PGWindowInitialize(PGWindow* pgw)
{
    pgw->dim=DIM2;  pgw->coord=CAR;
    pgw->margin=0.15;
    pgw->mag=1.0;  pgw->orgx=0;  pgw->orgy=0;
    pgw->x1min=0.0;  pgw->x1max=1.0;  pgw->x1size=1.0;
    pgw->x2min=0.0;  pgw->x2max=1.0;  pgw->x2size=1.0;
    pgw->x3min=0.0;  pgw->x3max=1.0;  pgw->x3size=1.0;
    pgw->phi=PI/6.0; pgw->theta=PI/6.0;
    pgw->x2_x1=0.0;  pgw->x3_x1=0.0;
}

PGWindow* CreatePGFrame(int gmx, int gmy, unsigned int wdx,
                        unsigned int wdy, char* title)
{
    XEvent e;
    PGWindow* pgw;
    XSizeHints sh;
    XSetWindowAttributes sat;
    pgw=(PGWindow*)malloc(sizeof(PGWindow));
    pgw->parent=&pgroot;
    pgw->pgwl.next=NULL;
    pgw->pgl.next=NULL;
    PGWindowInitialize(pgw);
    AddPGWindow(&pgroot, pgw);
    if(nowindow) return pgw;
    pgw->bg=BlackPixel(display, XDefaultScreen(display));
    pgw->w=XCreateSimpleWindow
            (display, DefaultRootWindow(display),
             gmx, gmy, wdx, wdy, 0, 0, pgw->bg);
    pgw->ev.type=FALSE;
    pgw->p=(Pixmap)NULL;
    pgw->wdx=wdx;  pgw->wdy=wdy;
    sh.flags=(USPosition);
    XSetWMNormalHints(display, pgw->w, &sh);
    XSelectInput(display, pgw->w, KeyPressMask
                 | ExposureMask
                 | StructureNotifyMask);
    sat.backing_store=Always;
    XChangeWindowAttributes(display, pgw->w, CWBackingStore, &sat);
    if(title != NULL){
        XStoreName(display, pgw->w, title);
        XSetIconName(display, pgw->w, title);
    }
    pgw->gc=XCreateGC(display, pgw->w, 0, 0);
    XMapWindow(display, pgw->w);
    XWindowEvent(display, pgw->w, StructureNotifyMask, &e);
    while(e.type!=MapNotify)
            XWindowEvent(display, pgw->w, StructureNotifyMask, &e);
    return pgw;
}

PGWindow* CreatePGWindow(PGWindow* parent,
                         double x, double y,
                         double wid, double hgt)
{
    XEvent e;
    PGWindow* pgw;
    XSetWindowAttributes sat;
    pgw=(PGWindow*)malloc(sizeof(PGWindow));
    pgw->parent=parent;
    pgw->pgwl.next=NULL;
    pgw->pgl.next=NULL;
    PGWindowInitialize(pgw);
    AddPGWindow(parent, pgw);
    if(nowindow) return pgw;
    pgw->bg=BlackPixel(display, XDefaultScreen(display));
    pgw->w=XCreateSimpleWindow
            (display, parent->w, 0, 0, parent->wdx, parent->wdy,
             0, 0, pgw->bg);
    pgw->ev.type=FALSE;
    pgw->p=(Pixmap)NULL;
    XSelectInput(display, pgw->w, StructureNotifyMask);
    sat.backing_store=Always;
    XChangeWindowAttributes(display, pgw->w, CWBackingStore, &sat);
    pgw->gc=XCreateGC(display, pgw->w, 0, 0);
    MoveResizePGWindow(pgw, x, y, wid, hgt);
    XMapWindow(display, pgw->w);
    XWindowEvent(display, pgw->w, StructureNotifyMask, &e);
    while(e.type!=MapNotify)
            XWindowEvent(display, pgw->w, StructureNotifyMask, &e);
    XSelectInput(display, pgw->w,
                 KeyPressMask
                 | ButtonPressMask
                 | ButtonMotionMask
                 | ButtonReleaseMask);
    return pgw;
}


/* show window */

void ShowPGWindow(PGWindow* pgw)
{
    PGObjectList* pgl;
    if(nowindow) return;
    if(pgw==&pgroot) return;
    if(pgw->p==(Pixmap)NULL){
        pgw->p=XCreatePixmap(display, pgw->w, pgw->wdx, pgw->wdy,
                             XDefaultDepth(display, 0));
    }
    XSetForeground(display, pgw->gc, pgw->bg);
    XFillRectangle(display, pgw->p, pgw->gc, 0, 0,
                   pgw->wdx, pgw->wdy);
    AxisVector(pgw, 'x', (double)(pgw->wdx), (double)(pgw->wdy));
    pgl=pgw->pgl.next;
    while(pgl){
        if(pgl->pg==NULL){
        }else if(pgl->pg->type==PGPOINTS){
            ShowPGPoints(pgw, (PGPoints*)(pgl->pg));
        }else if(pgl->pg->type==PGLINE){
            ShowPGLine(pgw, (PGLine*)(pgl->pg));
        }else if(pgl->pg->type==PGSTRING){
            ShowPGString(pgw, (PGString*)(pgl->pg));
        }else if(pgl->pg->type==PGFIXSTRING){
            ShowPGFixString(pgw, (PGFixString*)(pgl->pg));
        }else if(pgl->pg->type==PGARROWS){
            ShowPGArrows(pgw, (PGArrows*)(pgl->pg));
        }
        pgl=pgl->next;
    }
    XCopyArea(display, pgw->p, pgw->w, pgw->gc,
              0, 0, pgw->wdx, pgw->wdy, 0, 0);
    XFlush(display);
}
void ShowPGWindows(PGWindow* pgw)
{
    PGWindowList* pgwl;
    if(nowindow) return;
    ShowPGWindow(pgw);
    pgwl=pgw->pgwl.next;
    while(pgwl){
        ShowPGWindows(pgwl->pgw);
        pgwl=pgwl->next;
    }
}
void ShowAllPGWindows(){
    ShowPGWindows(&pgroot);
}

/* dump window */

#ifdef USE_LIB_XPM
int DumpPGWindowToXpm(PGWindow* pgw, char* fname)
{
    int ret;
    XImage* ximg;
    if(nowindow) return NoWindow;
    ximg=XGetImage(display, pgw->p, 0, 0, pgw->wdx, pgw->wdy,
                   AllPlanes, XYPixmap);
    ret=XpmWriteFileFromImage(display, fname, ximg, NULL, NULL);
    XDestroyImage(ximg);
    return ret;
}
#else
int DumpPGWindowToXpm(PGWindow* pgw, char* fname)
{
    return Unsuported;
}
#endif /* USE_LIB_XPM */

#if defined(USE_LIB_UNGIF) || defined(USE_LIB_GIF)
/* based on dev2gif.c in libungif library */
int HandleGifError(GifFileType* gfile)
{
    int ret;
    PrintGifError();
    ret=GifLastError();
    if(EGifCloseFile(gfile)==GIF_ERROR) GifLastError();
    return ret;
}
int DumpPGWindowToGif(PGWindow* pgw, char* fname)
{
    int i, j, ii, wdx, wdy;
    int scbits;
    unsigned long pix;
    XColor xclt[256];
    XImage* ximg;
    GifFileType* gfile;
    GifByteType* buffer;
    GifPixelType* line;
    GifColorType gcm256[256];
    ColorMapObject* gcm=NULL;
    if(nowindow) return NoWindow;
    wdx=pgw->wdx;  wdy=pgw->wdy;
    ximg=XGetImage(display, pgw->p, 0, 0, wdx, wdy,
                   AllPlanes, XYPixmap);
    buffer=(GifByteType*)malloc(sizeof(GifByteType)*wdx*wdy);
    for(i=0; i<256; i++) xclt[i].pixel=0;
    ii=0;
    for(j=0; j<wdy; j++){
    for(i=0; i<wdx; i++){
        pix=XGetPixel(ximg, i, j);
        if(255<pix) pix=255;
        xclt[pix].pixel=pix;
        buffer[ii]=(GifByteType)pix;  ii++;
    }}
    XQueryColors(display, cm, xclt, 256);
    gcm=MakeMapObject(256, gcm256);
    j=0;
    gcm->Colors[j].Red  =xclt[0].red  /256;
    gcm->Colors[j].Green=xclt[0].green/256;
    gcm->Colors[j].Blue =xclt[0].blue /256;
    xclt[0].pixel=j;  j++;
    for(i=1; i<256; i++){
        if(xclt[i].pixel){
            gcm->Colors[j].Red  =xclt[i].red  /256;
            gcm->Colors[j].Green=xclt[i].green/256;
            gcm->Colors[j].Blue =xclt[i].blue /256;
            xclt[i].pixel=j;  j++;
        }
    }
    for(scbits=1; (1<<scbits)<j && scbits<8; scbits++);
    for(ii=0; ii<wdx*wdy; ii++){
        buffer[ii]=xclt[(int)(buffer[ii])].pixel;
    }
    XDestroyImage(ximg);
    line=(GifPixelType*)malloc(sizeof(GifPixelType)*wdx);
    if(((gfile=EGifOpenFileName(fname, FALSE))==NULL)
       || (EGifPutScreenDesc(gfile, wdx, wdy, scbits, 0, gcm)
           ==GIF_ERROR)
       || (EGifPutImageDesc(gfile, 0, 0, wdx, wdy, FALSE, NULL)
           ==GIF_ERROR)){
        free(buffer);  free(line);
        return HandleGifError(gfile);
    }
    ii=0;
    for(j=0; j<wdy; j++){
        memcpy(line, buffer+ii, wdx*sizeof(GifPixelType));
        ii+=wdx;
        if(EGifPutLine(gfile, line, wdx)==GIF_ERROR){
            free(buffer);  free(line);
            return HandleGifError(gfile);
        }
    }
    if(EGifCloseFile(gfile)==GIF_ERROR){
        free(buffer);  free(line);
        return HandleGifError(gfile);
    }
    free(buffer);  free(line);
    return 0;
}
#else
int DumpPGWindowToGif(PGWindow* pgw, char* fname)
{
    return Unsuported;
}
#endif /* USE_LIB_UNGIF || USE_LIB_GIF */


/* raise */

void RaisePGFrame(PGWindow* pgw)
{
    PGWindow* parent;
    if(nowindow) return;
    parent=ParentPGFrame(pgw);
    XRaiseWindow(display, parent->w);
}

/* event handling */

void EventHandling(PGWindow* pgw)
{
    PGWindowList* pgwl;
    /* event handling */
    switch(pgw->ev.type){
      case Expose:
        ShowPGWindows(pgw);
        break;
      case ConfigureNotify:
        if(!IsPGFrame(pgw)) break;
        pgw->wdx=pgw->ev.xconfigure.width;
        pgw->wdy=pgw->ev.xconfigure.height;
        pgwl=pgw->pgwl.next;
        while(pgwl){
            ReconfigurePGWindow(pgwl->pgw);
            pgwl=pgwl->next;
        }
        break;
    }
    pgw->ev.type=FALSE;
}
void RotateEvent(PGWindow* pgw, XEvent e)
{
    double step=PI/36.0;
    static int px, py;
    int n;
    double ang1, ang2;
    if(!Is3D(pgw)) return;
    switch(e.type){
      case ButtonPress:
        px=e.xbutton.x;
        py=e.xbutton.y;
        break;
      case MotionNotify:
        n=XPending(display);
        if(n==0){
            step*=0.25;
        }else if(n/2*2==n){
            return;
        }
        ang1=(double)(px-e.xbutton.x)/pgw->wdx*PI;
        ang1=(int)(ang1/step)*step;
        if(ang1!=0.0){
            IncreasePhi(pgw, ang1);
            px=e.xbutton.x;
        }
        ang2=(double)(e.xbutton.y-py)/pgw->wdy*PI;
        ang2=(int)(ang2/step)*step;
        if(ang2!=0.0){
            IncreaseTheta(pgw, ang2);
            py=e.xbutton.y;
            ShowPGWindow(pgw);
        }else if(ang1!=0.0){
            ShowPGWindow(pgw);
        }
        break;
      case ButtonRelease:
        break;
    }
}
void MoveEvent(PGWindow* pgw, XEvent e)
{
    static int px, py;
    int n, dx, dy, step=3;
    switch(e.type){
      case ButtonPress:
        px=e.xbutton.x;
        py=e.xbutton.y;
        break;
      case MotionNotify:
        n=XPending(display);
        if(n==0){
            step=0;
        }else if(n/2*2==n){
            return;
        }
        dx=e.xbutton.x-px;
        if(dx<-step || step<dx){
            pgw->orgx+=dx/(double)pgw->wdx;
            px=e.xbutton.x;
        }
        dy=py-e.xbutton.y;
        if(dy<-step || step<dy){
            pgw->orgy+=dy/(double)pgw->wdy;
            py=e.xbutton.y;
            ShowPGWindow(pgw);
        }else if(dx<-step || step<dx){
            ShowPGWindow(pgw);
        }
        break;
      case ButtonRelease:
        break;
    }
}
void PGWMouseEventHandling(XEvent e)
{
    static int btn;
    PGWindow* pgw;
    if(!(pgw=XWindowToPGWindow(&pgroot, e.xbutton.window))) return;
    if(e.type==ButtonPress) btn=e.xbutton.button;
    if(btn==1){
        RotateEvent(pgw, e);
    }else if(btn==2){
        MoveEvent(pgw, e);
    }
}


void HandleFrames()
{
    PGWindowList* pgwl;
    pgwl=pgroot.pgwl.next;
    while(pgwl){
        EventHandling(pgwl->pgw);
        pgwl=pgwl->next;
    }
}

int PGWWaitEvent(XEvent* e)
{
    Window w;
    PGWindow* pgw;
    if(nowindow) return FALSE;
    while(TRUE){
        XNextEvent(display, e);
        if(e->type==KeyPress
           || e->type==ButtonPress
           || e->type==ButtonRelease
           || e->type==MotionNotify){
            HandleFrames();
            return TRUE;
        }else if(e->type==Expose){
            w=e->xexpose.window;
        }else if(e->type==ConfigureNotify){
            w=e->xconfigure.window;
        }else{
            if(!XPending(display)) HandleFrames();
            continue;
        }
        pgw=XWindowToPGWindow(&pgroot, w);
        if(IsPGFrame(pgw) && (pgw->ev.type==FALSE
               || e->type==ConfigureNotify)) pgw->ev=*e;
        if(!XPending(display)) HandleFrames();
    }
}
int PGWWaitKeyPress(XEvent* e)
{
    Window w;
    PGWindow* pgw;
    if(nowindow) return FALSE;
    while(TRUE){
        XNextEvent(display, e);
        if(e->type==KeyPress){
            HandleFrames();
            return TRUE;
        }else if(e->type==ButtonPress
                 || e->type==ButtonRelease
                 || e->type==MotionNotify){
            PGWMouseEventHandling(*e);
            continue;
        }else if(e->type==Expose){
            w=e->xexpose.window;
        }else if(e->type==ConfigureNotify){
            w=e->xconfigure.window;
        }else{
            if(!XPending(display)) HandleFrames();
            continue;
        }
        pgw=XWindowToPGWindow(&pgroot, w);
        if(IsPGFrame(pgw) && (pgw->ev.type==FALSE
               || e->type==ConfigureNotify)) pgw->ev=*e;
        if(!XPending(display)) HandleFrames();
    }
}

int PGWPendingEvent(XEvent* e)
{
    Window w;
    PGWindow* pgw;
    if(nowindow) return FALSE;
    while(XPending(display)){
        XNextEvent(display, e);
        if(e->type==KeyPress
           || e->type==ButtonPress
           || e->type==ButtonRelease
           || e->type==MotionNotify){
            HandleFrames();
            return TRUE;
        }else if(e->type==Expose){
            w=e->xexpose.window;
        }else if(e->type==ConfigureNotify){
            w=e->xconfigure.window;
        }else{
            continue;
        }
        pgw=XWindowToPGWindow(&pgroot, w);
        if(IsPGFrame(pgw) && (pgw->ev.type==FALSE
               || e->type==ConfigureNotify)) pgw->ev=*e;
    }
    HandleFrames();
    return FALSE;
}
int PGWPendingKeyPress(XEvent* e)
{
    Window w;
    PGWindow* pgw;
    if(nowindow) return FALSE;
    while(XPending(display)){
        XNextEvent(display, e);
        if(e->type==KeyPress){
            HandleFrames();
            return TRUE;
        }else if(e->type==ButtonPress
                 || e->type==ButtonRelease
                 || e->type==MotionNotify){
            PGWMouseEventHandling(*e);
            continue;
        }else if(e->type==Expose){
            w=e->xexpose.window;
        }else if(e->type==ConfigureNotify){
            w=e->xconfigure.window;
        }else{
            continue;
        }
        pgw=XWindowToPGWindow(&pgroot, w);
        if(IsPGFrame(pgw) && (pgw->ev.type==FALSE
               || e->type==ConfigureNotify)) pgw->ev=*e;
    }
    HandleFrames();
    return FALSE;
}
