#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <X11/Xlib.h>
#include <math.h>
#include "pgw.h"

extern Display* display;




double TransScale(double wid, double hgt, double r)
{
    double a, b, a4wid, a4hgt, ret;
    a=wid*wid+hgt*hgt;
    a4wid=A4LONG*(1.0-2.0*PSMARGIN);
    a4hgt=A4SHORT*(1.0-2.0*PSMARGIN);
    b=(double)(a4wid*a4wid+a4hgt*a4hgt);
    ret=sqrt(a/b)*r;
    return ret;
}

void SetLineAttributes(PGWindow* pgw, PGP pgp)
{
    char dash[8];
    unsigned int width;
    double linewidth;
    linewidth=pgp.width*BaseLineWidth()*PSLINEWIDTH;
    if(linewidth<=0.0){
        width=0;
    }else{
        linewidth=TransScale((double)(pgw->wdx), (double)(pgw->wdy),
                             linewidth);
        width=(unsigned int)(linewidth+0.5);
        if(width==1) width=0;
    }
    if(pgp.ltype==DOT){
        dash[0]=2;  dash[1]=2*3;
        XSetDashes(display, pgw->gc, 0, dash, 2);
        XSetLineAttributes(display, pgw->gc, width,
                           LineOnOffDash, CapRound, JoinBevel);
    }else if(pgp.ltype==BRK){
        double len;
        len=TransScale((double)(pgw->wdx), (double)(pgw->wdy),
                       BRKLINELEN);
        dash[0]=len;  dash[1]=len;
        XSetDashes(display, pgw->gc, 0, dash, 2);
        XSetLineAttributes(display, pgw->gc, width,
                           LineOnOffDash, CapRound, JoinBevel);
    }else{
        XSetLineAttributes(display, pgw->gc, width,
                           LineSolid, CapRound, JoinBevel);
    }
}

/* draw X string */

XFontStruct* GetXFont(double p, int fmly, int sl)
{
    int ipt=0;  int pt;
    static int getfont_init=FALSE;
    static char fontname[FMLY][SL][MAX_STR]=FONTNAMES;
    static XFontStruct* fontset[MAXPT+1][FMLY][SL];
    static XFontStruct* oldfs;
    if(!getfont_init){
        int i, j, k;
        for(k=0; k<MAXPT+1; k++){
            for(j=0; j<FMLY; j++){
                for(i=0; i<SL; i++) fontset[k][j][i]=NULL;
            }
        }
        getfont_init=TRUE;
    }
    if(p<MINPT){
        p=MINPT;
    }else if(MAXPT<p){
        p=MAXPT;
    }
    pt=(int)(p/PTSTEP+0.5)*PTSTEP;
    while(fontset[pt][fmly][sl]==NULL){
        char fontstr[MAX_STR];
        char spt[MAX_STR];
        if(pt+ipt<MINPT){
            ipt=-ipt;  ipt+=PTSTEP;
        }
        sprintf(spt, "%d", pt+ipt);
        strcpy(fontstr, "-*-");
        strcat(fontstr, fontname[fmly][sl]);
        strcat(fontstr, "-*-*-");
        strcat(fontstr, spt);
        strcat(fontstr, "-*-*-*-*-*-*-*");
        fontset[pt+ipt][fmly][sl]=
                XLoadQueryFont(display, fontstr);
        if(fontset[pt+ipt][fmly][sl]){
            fontset[pt][fmly][sl]=fontset[pt+ipt][fmly][sl];
            if(fmly==SYMB)
                    fontset[pt][fmly][1-sl]=fontset[pt][fmly][sl];
        }else{
            if(ipt>0){
                ipt=-ipt;
            }else{
                ipt=-ipt;  ipt+=PTSTEP;
                if(pt+ipt>MAXPT) return oldfs;
            }
        }
    }
    oldfs=fontset[pt][fmly][sl];
    return fontset[pt][fmly][sl];
}

void DrawXCharacters(PGWindow* pgw, int pre, int* px, int* py,
                     int rot, char* str, int l)
{
    unsigned int w, h;
    int dir, asc, des, sub;
    XCharStruct ov;
    XFontStruct* ftst;
    PGFontList fl;
    if(l<1) return;
    fl=CurrentPGFont();
    ftst=GetXFont(fl.pt, fl.fmly, fl.sl);
    sub=(int)(fl.sub+0.5);
    XSetFont(display, pgw->gc, ftst->fid);
    XTextExtents(ftst, str, l, &dir, &asc, &des, &ov);
    w=(unsigned int)(ov.width);
    h=(unsigned int)(ov.ascent);
    if(rot==R0){
        if(!pre){
            XDrawString(display, pgw->p, pgw->gc,
                        *px, *py-sub, str, l);
        }
        (*px)+=(int)w;
    }else if(rot==R90){
        if(!pre){
            XDrawRectangle(display, pgw->p, pgw->gc,
                           (int)(*px-h-sub), (int)(*py-w), h, w);
        }
        (*py)-=(int)w;
    }else if(rot==R180){
        if(!pre){
            XDrawRectangle(display, pgw->p, pgw->gc,
                           (int)(*px-w), *py+sub, w, h);
        }
        (*px)-=(int)w;
    }else if(rot==R270){
        if(!pre){
            XDrawRectangle(display, pgw->p, pgw->gc,
                           *px+sub, *py, h, w);
        }
        (*py)+=(int)w;
    }
}

int LabelCommand(char* str, int* n)
{
    int ret=TRUE;
    int k;
    char buf[256];
    switch(str[0]){
      case 'r':
        SetSlant(RM);
        break;
      case 'i':
        SetSlant(IT);
        break;
      case 'c':
        SetFontFamily(COUR);
        break;
      case 't':
        SetFontFamily(TIMS);
        break;
      case 'h':
        SetFontFamily(HELV);
        break;
      case 's':
        SetFontFamily(SYMB);
        break;
      case 'x':
        k=0;
        while(isdigit(str[k+1]) || str[k+1]=='.'){
            buf[k]=str[k+1];  k++;
        }
        if(k){
            buf[k]='\0';
            MultipleFontPoint(atof(buf));
            (*n)+=k;
        }
        break;
      case '^':
        AddSubscript(0.8*CurrentFontPoint());
        MultipleFontPoint(SUBPT);
        AddSubscript(-0.5*CurrentFontPoint());
        break;
      case '_':
        MultipleFontPoint(SUBPT);
        AddSubscript(-0.5*CurrentFontPoint());
        break;
      default:
        ret=FALSE;
    }
    (*n)++;
    return ret;
}




void DrawXPGString(PGWindow* pgw, char* string, int flush, int rot,
                   double pt, int px, int py)
{
    static int pre=FALSE;
    static int dx, dy;
    int pxc, pyc;
    int i=0;  int j=0;
    char str[MAX_STR];
    char buf[MAX_STR];
    dx=0;  dy=0;
    if(!pre && flush!=RIGHT){
        pre=TRUE;
        DrawXPGString(pgw, string, flush, rot, pt, px, py);
        pre=FALSE;
        if(flush==CENTER){
            px-=0.5*dx;  py-=0.5*dy;
        }else if(flush==LEFT){
            px-=dx;  py-=dy;
        }
    }
    pxc=px;  pyc=py;
    strcpy(str, string);
    PushPGFontList(pt, DefaultFontFamily(), DefaultSlant(), 0.0);
    while(str[i]!='\0'){
        switch(str[i]){
          case 123:
            buf[j]='\0';
            DrawXCharacters(pgw, pre, &pxc, &pyc, rot, buf, j);
            j=0;  i++;
            DuplicateFontList();
            break;
          case 125:
            buf[j]='\0';
            DrawXCharacters(pgw, pre, &pxc, &pyc, rot, buf, j);
            j=0;  i++;
            PopPGFontList();
            break;
          case 33:
          case 92:
            buf[j]='\0';
            DrawXCharacters(pgw, pre, &pxc, &pyc, rot, buf, j);
            j=0;  i++;
            if(!LabelCommand(str+i, &i)){
                buf[j]=str[i];  j++;  i++;
            }
          default:
            buf[j]=str[i];
            j++;  i++;
        }
    }
    buf[j]='\0';
    DrawXCharacters(pgw, pre, &pxc, &pyc, rot, buf, j);
    ClearFontList();
    if(pre){
        dx=pxc-px;  dy=pyc-py;
    }
}


/* transformation */

double* LimitLocate(int n, double* x, double min, double max)
{
    int i;
    double dmin, dmax;
    double* ret;
    ret=(double*)malloc(n*sizeof(double));
    if(x==NULL){
        for(i=0; i<n; i++) ret[i]=min;
        return ret;
    }
    dmin=min-5*(max-min);  dmax=max+5*(max-min);
    for(i=0; i<n; i++){
        if(x[i]<dmin){
            ret[i]=dmin;
        }else if(dmax<x[i]){
            ret[i]=dmax;
        }else{
            ret[i]=x[i];
        }
    }
    return ret;
}
void D2CarAxis(PGWindow* pgw, char dev, double wdx, double wdy)
{
    double xmargin, ymargin;
    double wid, hgt;
    double mx, my;
    xmargin=wdx*pgw->margin;  ymargin=wdy*pgw->margin;
    wid=wdx-2.0*xmargin;  hgt=wdy-2.0*ymargin;
    if(pgw->x2_x1>0.0){
	if(hgt/wid>pgw->x2_x1){
	    hgt=wid*pgw->x2_x1;
	}else{
	    wid=hgt/pgw->x2_x1;
	}
    }
    mx=pgw->orgx*wdx;  my=pgw->orgy*wdy;
    if(dev=='x'){
        hgt=-hgt;  my=-my;
    }
    pgw->x1vecx=wid*pgw->mag;  pgw->x2vecy=hgt*pgw->mag;
    pgw->zerox=0.5*(wdx-pgw->x1vecx)+mx;
    pgw->zeroy=0.5*(wdy-pgw->x2vecy)+my;
}
void D2CarTrans(PGWindow* pgw, double x1vecx, double x2vecy,
                double zerox, double zeroy, double* x1, double* x2,
                double* px, double* py, int num)
{
    int i;
    double tmp1, tmp2;
    double* xx1;  double* xx2;
    xx1=LimitLocate(num, x1, pgw->x1min, pgw->x1max);
    xx2=LimitLocate(num, x2, pgw->x2min, pgw->x2max);
    tmp1=x1vecx/pgw->x1size;
    tmp2=x2vecy/pgw->x2size;
    for(i=0; i<num; i++){
        px[i]=zerox+(xx1[i]-pgw->x1min)*tmp1;
        py[i]=zeroy+(xx2[i]-pgw->x2min)*tmp2;
    }
    free(xx1);  free(xx2);
}
void D3CarAxis(PGWindow* pgw, char dev, double wdx, double wdy)
{
    double x2_x1=1.0;  double x3_x1=wdy/wdx;
    double width, height;
    double phi, sint, alph;
    double xmargin;
    double xx, xy;
    double mx, my;
    if(pgw->x2_x1>0.0) x2_x1=pgw->x2_x1;
    if(pgw->x3_x1>0.0) x3_x1=pgw->x3_x1;
    phi=pgw->phi;  sint=sin(pgw->theta);  alph=atan(x2_x1);
    xmargin=D3MARGIN*wdx*pgw->margin;
    width=wdx-2.0*xmargin;
    xx=0.5*width*(cos(0.25*PI+alph)+cos(0.25*PI-alph));
    xy=0.5*width*(sin(0.25*PI+alph)+sin(0.25*PI-alph))*0.5;
    height=x3_x1*sqrt(xx*xx+xy*xy);
    pgw->zerox=xmargin+0.5*width*(1.0-cos(phi-alph));
    pgw->zeroy=0.5*(wdy-height+width*sin(phi-alph)*sint);
    pgw->x1vecx= 0.5*width*(cos(phi+alph)+cos(phi-alph));
    pgw->x1vecy=-0.5*width*(sin(phi+alph)+sin(phi-alph))*sint;
    pgw->x2vecx= 0.5*width*(cos(phi-alph)-cos(phi+alph));
    pgw->x2vecy=-0.5*width*(sin(phi-alph)-sin(phi+alph))*sint;
/*    pgw->x3vecx=0.0;*/
    pgw->x3vecy=height;
    mx=pgw->orgx*wdx;  my=pgw->orgy*wdy;
    if(dev=='x'){
        pgw->zeroy =wdy-pgw->zeroy;  pgw->x1vecy=-(pgw->x1vecy);
        pgw->x2vecy=-(pgw->x2vecy);  pgw->x3vecy=-(pgw->x3vecy);
        my=-my;
    }
    pgw->x1vecx*=pgw->mag;  pgw->x1vecy*=pgw->mag;
    pgw->x2vecx*=pgw->mag;  pgw->x2vecy*=pgw->mag;
/*    pgw->x3vecx*=pgw->mag;*/  pgw->x3vecy*=pgw->mag;
    pgw->zerox+=mx;  pgw->zeroy+=my;
}
void D3CarTrans(PGWindow* pgw, double x1vecx, double x1vecy,
                double x2vecx, double x2vecy, double x3vecy,
                double zerox, double zeroy,
                double* x1, double* x2, double* x3,
                double* px, double* py, int num)
{
    int i;
    double tmp1, tmp2, tmp3, tmp4, tmp5, tmp6;
    double* xx1;  double* xx2;  double* xx3;
    xx1=LimitLocate(num, x1, pgw->x1min, pgw->x1max);
    xx2=LimitLocate(num, x2, pgw->x2min, pgw->x2max);
    xx3=LimitLocate(num, x3, pgw->x3min, pgw->x3max);
    tmp1=x1vecx/pgw->x1size;
    tmp2=x2vecx/pgw->x2size;
/*    tmp3=x3vecx/pgw->x3size;*/
    tmp4=x1vecy/pgw->x1size;
    tmp5=x2vecy/pgw->x2size;
    tmp6=x3vecy/pgw->x3size;
    for(i=0; i<num; i++){
        px[i]=zerox+(xx1[i]-pgw->x1min)*tmp1
                   +(xx2[i]-pgw->x2min)*tmp2;
/*                   +(xx2[i]-pgw->x3min)*tmp3;*/
        py[i]=zeroy+(xx1[i]-pgw->x1min)*tmp4
                   +(xx2[i]-pgw->x2min)*tmp5
                   +(xx3[i]-pgw->x3min)*tmp6;
    }
    free(xx1);  free(xx2);  free(xx3);
}
void D2CylAxis(PGWindow* pgw, char dev, double wdx, double wdy)
{
    double xmargin, ymargin;
    double wid, hgt;
    double mx, my;
    xmargin=wdx*pgw->margin;  ymargin=wdy*pgw->margin;
    wid=0.5*wdx-xmargin;  hgt=0.5*wdy-ymargin;
    if(wid > hgt){
        pgw->radius=hgt;
    }else{
        pgw->radius=wid;
    }
    mx=pgw->orgx*wdx;  my=pgw->orgy*wdy;
    if(dev=='x') my=-my;
    pgw->zerox=0.5*wdx;  pgw->zeroy=0.5*wdy;
    pgw->radius*=pgw->mag;
    pgw->zerox+=mx;  pgw->zeroy+=my;
}
void D2CylTrans(PGWindow* pgw, char dev, double radius,
                double zerox, double zeroy, double* x1, double* x2,
                double* px, double* py, int num)
{
    int i;
    double tmp1, tmp2;
    double* xx1;  double* xx2;
    double rad, phi;
    xx1=LimitLocate(num, x1, pgw->x1min, pgw->x1max);
    xx2=LimitLocate(num, x2, pgw->x2min, pgw->x2max);
    tmp1=radius/pgw->x1size;
    tmp2=2*PI/pgw->x2size;
    for(i=0; i<num; i++){
        rad=(xx1[i]-pgw->x1min)*tmp1;
        phi=(xx2[i]-pgw->x2min)*tmp2;
        if(dev=='x') phi=-phi;
        px[i]=zerox+rad*cos(phi);
        py[i]=zeroy+rad*sin(phi);
    }
    free(xx1);  free(xx2);
}
void D3CylAxis(PGWindow* pgw, char dev, double wdx, double wdy)
{
    double x3_x1=wdy/wdx;
    double xmargin;
    double mx, my;
    if(pgw->x3_x1>0.0) x3_x1=pgw->x3_x1;
    xmargin=D3MARGIN*wdx*pgw->margin;
    pgw->radius=0.5*wdx-xmargin;
    pgw->height=pgw->radius*x3_x1;
    mx=pgw->orgx*wdx;  my=pgw->orgy*wdy;
    if(dev=='x'){
        pgw->x3vecy=-(pgw->height);  my=-my;
    }
    pgw->radius*=pgw->mag;  pgw->x3vecy*=pgw->mag;
    pgw->zerox=0.5*wdx;  pgw->zeroy=0.5*(wdy-pgw->x3vecy);
    pgw->zerox+=mx;  pgw->zeroy+=my;
}
void D3CylTrans(PGWindow* pgw, char dev,  double radius,
                double x3vecy, double zerox, double zeroy,
                double* x1, double* x2, double* x3,
                double* px, double* py, int num)
{
    int i;
    double* xx1;  double* xx2;  double* xx3;
    double sint, rad, phi;
    double tmp1, tmp2, tmp3;
    xx1=LimitLocate(num, x1, pgw->x1min, pgw->x1max);
    xx2=LimitLocate(num, x2, pgw->x2min, pgw->x2max);
    xx3=LimitLocate(num, x3, pgw->x3min, pgw->x3max);
    sint=sin(pgw->theta);
    tmp1=radius/pgw->x1size;
    tmp2=2*PI/pgw->x2size;
    tmp3=x3vecy/pgw->x3size;
    for(i=0; i<num; i++){
        rad=(xx1[i]-pgw->x1min)*tmp1;
        phi=(xx2[i]-pgw->x2min)*tmp2;
        phi-=pgw->phi;
        if(dev=='x') phi=-phi;
        px[i]=zerox+rad*cos(phi);
        py[i]=zeroy+rad*sint*sin(phi)
                +(xx3[i]-pgw->x3min)*tmp3;
    }
    free(xx1);  free(xx2);  free(xx3);
}

int AxisVector(PGWindow* pgw, char dev, double wdx, double wdy)
{
    if(pgw->coord==CAR){
        switch(pgw->dim){
          case DIM2:
            D2CarAxis(pgw, dev, wdx, wdy);
            break;
          case DIM3:
            D3CarAxis(pgw, dev, wdx, wdy);
            break;
          default:
            return FALSE;
        }
    }else if(pgw->coord==CYL){
        switch(pgw->dim){
          case DIM2:
            D2CylAxis(pgw, dev, wdx, wdy);
            break;
          case DIM3:
            D3CylAxis(pgw, dev, wdx, wdy);
            break;
          default:
            return FALSE;
        }
    }else{
        return FALSE;
    }
    return TRUE;
}
int TransCoords(PGWindow* pgw, char dev,
                double* x1, double* x2, double* x3,
                double* px, double* py, int num)
{
    if(pgw->coord==CAR){
        switch(pgw->dim){
          case DIM2:
            D2CarTrans(pgw, pgw->x1vecx, pgw->x2vecy,
                       pgw->zerox, pgw->zeroy,
                       x1, x2, px, py, num);
            break;
          case DIM3:
            D3CarTrans(pgw, pgw->x1vecx, pgw->x1vecy,
                       pgw->x2vecx, pgw->x2vecy, pgw->x3vecy,
                       pgw->zerox, pgw->zeroy,
                       x1, x2, x3, px, py, num);
            break;
          default:
            return FALSE;
        }
    }else if(pgw->coord==CYL){
        switch(pgw->dim){
          case DIM2:
            D2CylTrans(pgw, dev, pgw->radius,
                       pgw->zerox, pgw->zeroy,
                       x1, x2, px, py, num);
            break;
          case DIM3:
            D3CylTrans(pgw, dev, pgw->radius, pgw->x3vecy,
                       pgw->zerox, pgw->zeroy,
                       x1, x2, x3, px, py, num);
            break;
          default:
            return FALSE;
        }
    }else{
        return FALSE;
    }
    return TRUE;
}

void ShowPGPoints(PGWindow* pgw, PGPoints* pg)
{
    int i;
    XPoint* pts;
    double* px;  double* py;
    double rad;
    px=(double*)malloc(pg->num*sizeof(double));
    py=(double*)malloc(pg->num*sizeof(double));
    if(!(TransCoords(pgw, 'x',
                     pg->x1, pg->x2, pg->x3, px, py, pg->num))){
        free(px);  free(py);
        return;
    }
    pts=(XPoint*)malloc(pg->num*sizeof(XPoint));
    for(i=0; i<pg->num; i++){
        pts[i].x=(short)px[i];  pts[i].y=(short)py[i];
    }
    XSetForeground(display, pgw->gc, pg->pgp.pgcl.xcolor.pixel);
    SetLineAttributes(pgw, pg->pgp);
    rad=TransScale((double)(pgw->wdx), (double)(pgw->wdy),
                   pg->pgp.rad);
    if(pg->pgp.ptype==PTS){
        XDrawPoints(display, pgw->p, pgw->gc, pts, pg->num,
                    CoordModeOrigin);
    }else if(pg->pgp.ptype==SQR || pg->pgp.ptype==FSQR){
        XRectangle* rects;
        rects=(XRectangle*)malloc(pg->num*sizeof(XRectangle));
        for(i=0; i<pg->num; i++){
            rects[i].x=pts[i].x-(short)(0.5*rad*SQRTPI);
            rects[i].y=pts[i].y-(short)(0.5*rad*SQRTPI);
            rects[i].width =(unsigned short)(rad*SQRTPI);
            rects[i].height=(unsigned short)(rad*SQRTPI);
        }
        if(pg->pgp.ptype==FSQR){
            XFillRectangles(display, pgw->p, pgw->gc,
                            rects, pg->num);
        }else{
            XDrawRectangles(display, pgw->p, pgw->gc,
                            rects, pg->num);
        }
        free(rects);
    }else{
        XArc* arcs;
        arcs=(XArc*)malloc(pg->num*sizeof(XArc));
        for(i=0; i<pg->num; i++){
            arcs[i].x=pts[i].x-rad;  arcs[i].y=pts[i].y-rad;
            arcs[i].width =(unsigned short)(2.0*rad);
            arcs[i].height=(unsigned short)(2.0*rad);
            arcs[i].angle1=0;  arcs[i].angle2=360*64;
        }
        if(pg->pgp.ptype==FCIR){
            XFillArcs(display, pgw->p, pgw->gc, arcs, pg->num);
        }else{
            XDrawArcs(display, pgw->p, pgw->gc, arcs, pg->num);
        }
        free(arcs);
    }
    free(px);  free(py);  free(pts);
}

void ShowPGLine(PGWindow* pgw, PGLine* pg)
{
    int i;
    XPoint* pts;
    double* px;  double* py;
    px=(double*)malloc(pg->num*sizeof(double));
    py=(double*)malloc(pg->num*sizeof(double));
    if(!(TransCoords(pgw, 'x',
                     pg->x1, pg->x2, pg->x3, px, py, pg->num))){
        free(px);  free(py);
        return;
    }
    pts=(XPoint*)malloc(pg->num*sizeof(XPoint));
    for(i=0; i<pg->num; i++){
        pts[i].x=(short)px[i];  pts[i].y=(short)py[i];
    }
    XSetForeground(display, pgw->gc, pg->pgp.pgcl.xcolor.pixel);
    SetLineAttributes(pgw, pg->pgp);
    XDrawLines(display, pgw->p, pgw->gc, pts, pg->num,
               CoordModeOrigin);
    free(px);  free(py);  free(pts);
}

void ShowPGString(PGWindow* pgw, PGString* pg)
{
    double px, py, pt;
    if(!TransCoords(pgw, 'x',
                    &(pg->x1), &(pg->x2), &(pg->x3), &px, &py, 1))
            return;
    XSetForeground(display, pgw->gc, pg->pgp.pgcl.xcolor.pixel);
    pt=TransScale((double)(pgw->wdx), (double)(pgw->wdy),
                  pg->pgp.pt);
    DrawXPGString(pgw, pg->str, pg->pgp.flush,
                  pg->pgp.rot, pt, (int)px, (int)py);
}

void ShowPGFixString(PGWindow* pgw, PGFixString* pg)
{
    double px, py, pt;
    XSetForeground(display, pgw->gc, pg->pgp.pgcl.xcolor.pixel);
    pt=TransScale((double)(pgw->wdx), (double)(pgw->wdy),
                  pg->pgp.pt);
    px=pgw->wdx*pg->x;  py=pgw->wdy*pg->y;
    DrawXPGString(pgw, pg->str, pg->pgp.flush, pg->pgp.rot, pt,
                  (int)px, (int)py);
}

void ShowPGArrows(PGWindow* pgw, PGArrows* pg)
{
    int i;
    XSegment* sgs;
    double ex=pg->exp;
    double arrow=pg->pgp.arrow;  double angle=pg->pgp.angle;
    double a11, a12, a21, a22, vx, vy;
    double* px1;  double* py1;  double* px2;  double* py2;
    px1=(double*)malloc(pg->num*sizeof(double));
    py1=(double*)malloc(pg->num*sizeof(double));
    px2=(double*)malloc(pg->num*sizeof(double));
    py2=(double*)malloc(pg->num*sizeof(double));
    if(!(TransCoords(pgw, 'x',
                     pg->x1, pg->x2, pg->x3, px1, py1, pg->num))
    || !(TransCoords(pgw, 'x',
                     pg->v1, pg->v2, pg->v3, px2, py2, pg->num))){
        free(px1);  free(py1);  free(px2);  free(py2);
        return;
    }
    sgs=(XSegment*)malloc(pg->num*sizeof(XSegment));
    XSetForeground(display, pgw->gc, pg->pgp.pgcl.xcolor.pixel);
    SetLineAttributes(pgw, pg->pgp);
    for(i=0; i<pg->num; i++){
        sgs[i].x1=(short)px1[i];  sgs[i].y1=(short)py1[i];
        if(pg->exp<=0.0){
            double rad, vl;
            rad=TransScale((double)(pgw->wdx), (double)(pgw->wdy),
                           UNITARROW);
            vl=sqrt((px2[i]-px1[i])*(px2[i]-px1[i])
                    +(py2[i]-py1[i])*(py2[i]-py1[i]));
            ex=rad/vl;
        }
        px2[i]=px1[i]+ex*(px2[i]-px1[i]);
        py2[i]=py1[i]+ex*(py2[i]-py1[i]);
        sgs[i].x2=(short)px2[i];  sgs[i].y2=(short)py2[i];
    }
    XDrawSegments(display, pgw->p, pgw->gc, sgs, pg->num);
    a11=arrow*cos(angle);  a12=-arrow*sin(angle);
    a21=arrow*sin(angle);  a22= arrow*cos(angle);
    for(i=0; i<pg->num; i++){
        vx=px1[i]-px2[i];  vy=py1[i]-py2[i];
        sgs[i].x1=(short)(px2[i]+a11*vx+a12*vy);
        sgs[i].y1=(short)(py2[i]+a21*vx+a22*vy);
    }
    XDrawSegments(display, pgw->p, pgw->gc, sgs, pg->num);
    a11=arrow*cos(-angle);  a12=-arrow*sin(-angle);
    a21=arrow*sin(-angle);  a22= arrow*cos(-angle);
    for(i=0; i<pg->num; i++){
        vx=px1[i]-px2[i];  vy=py1[i]-py2[i];
        sgs[i].x1=(short)(px2[i]+a11*vx+a12*vy);
        sgs[i].y1=(short)(py2[i]+a21*vx+a22*vy);
    }
    XDrawSegments(display, pgw->p, pgw->gc, sgs, pg->num);
    free(px1);  free(py1);  free(px2);  free(py2);
    free(sgs);
}
